import logging
from functools import partial
from itertools import repeat, zip_longest
from pathlib import Path
from typing import List, Optional, Union
from warnings import warn
from weakref import ProxyType, proxy

import numpy as np
import pandas as pd

from ._config import config
from .backend import BACKENDS, FIELDS, ItemType
from .model import get_model
from .utils import as_str_list, check_year, parse_url, year_list

log = logging.getLogger(__name__)

#: Partial dimensions for “IAMC format”.
IAMC_IDX: List[Union[str, int]] = ["model", "scenario", "region", "variable", "unit"]


class Platform:
    """Instance of the modeling platform.

    A Platform connects two key components:

    1. A **back end** for storing data such as model inputs and outputs.
    2. One or more **model(s)**; codes in Python or other languages or frameworks that
       run, via :meth:`Scenario.solve`, on the data stored in the Platform.

    The Platform parameters control these components. :class:`TimeSeries` and
    :class:`Scenario` objects tied to a single Platform; to move data between platforms,
    see :meth:`Scenario.clone`.

    Parameters
    ----------
    name : str
        Name of a specific :ref:`configured <configuration>` backend.
    backend : 'jdbc'
        Storage backend type. 'jdbc' corresponds to the built-in
        :class:`.JDBCBackend`; see :obj:`.BACKENDS`.
    backend_args
        Keyword arguments to specific to the `backend`. See
        :class:`.JDBCBackend`.
    """

    # Storage back end for the platform
    _backend = None

    # List of method names which are handled directly by the backend
    _backend_direct = [
        "add_model_name",
        "add_scenario_name",
        "close_db",
        "get_doc",
        "get_meta",
        "get_model_names",
        "get_scenario_names",
        "open_db",
        "remove_meta",
        "set_doc",
        "set_meta",
    ]

    def __init__(self, name=None, backend=None, **backend_args):
        if name is None:
            if backend is None and not len(backend_args):
                # No arguments given: use the default platform config
                name = "default"
            elif backend is None:
                # Only backend_args given
                log.info("Using default JDBC backend")
                kwargs = {"class": "jdbc"}
            else:
                # Backend and maybe backend_args were given
                kwargs = {"class": backend}

        if name:
            # Using a named platform config; retrieve it
            self.name, kwargs = config.get_platform_info(name)

        # Overwrite any platform config with explicit keyword arguments
        kwargs.update(backend_args)

        # Retrieve the Backend class
        try:
            backend_class = kwargs.pop("class")
            backend_class = BACKENDS[backend_class]
        except KeyError:
            raise ValueError(
                f"backend class {repr(backend_class)} not among "
                + str(sorted(BACKENDS.keys()))
            )

        # Instantiate the backend
        self._backend = backend_class(**kwargs)

    def __getattr__(self, name):
        """Convenience for methods of Backend."""
        if name in self._backend_direct:
            return getattr(self._backend, name)
        else:
            raise AttributeError(name)

    def set_log_level(self, level):
        """Set log level for the Platform and its storage :class:`.Backend`.

        Parameters
        ----------
        level : str
            Name of a :py:ref:`Python logging level <levels>`.
        """
        if not (level in dir(logging) or isinstance(level, int)):
            raise ValueError(
                f"{repr(level)} is not a valid Python logger level, see "
                "https://docs.python.org/3/library/logging.html#logging-level"
            )

        # Set the level for the 'ixmp' logger
        # NB this may produce unexpected results when multiple Platforms exist and
        #    different log levels are set. To fix, could use a sub-logger per Platform
        #    instance.
        logging.getLogger("ixmp").setLevel(level)

        # Set the level for the 'ixmp.backend.*' logger. For JDBCBackend, this also has
        # the effect of setting the level for Java log messages that are printed to
        # stdout.
        self._backend.set_log_level(level)

    def get_log_level(self):
        """Return log level of the storage :class:`.Backend`, if any.

        Returns
        -------
        str
            Name of a :py:ref:`Python logging level <levels>`.
        """
        return self._backend.get_log_level()

    def scenario_list(self, default=True, model=None, scen=None):
        """Return information about TimeSeries and Scenarios on the Platform.

        Parameters
        ----------
        default : bool, optional
            Return *only* the default version of each TimeSeries/Scenario (see
            :meth:`TimeSeries.set_as_default`). Any (`model`, `scenario`)
            without a default version is omitted. If :obj:`False`, return all
            versions.
        model : str, optional
            A model name. If given, only return information for *model*.
        scen : str, optional
            A scenario name. If given, only return information for *scen*.

        Returns
        -------
        :class:`pandas.DataFrame`
            Scenario information, with the columns:

            - ``model``, ``scenario``, ``version``, and ``scheme``—Scenario
              identifiers; see :class:`Scenario`.
            - ``is_default``—:obj:`True` if the ``version`` is the default
              version for the (``model``, ``scenario``).
            - ``is_locked``—:obj:`True` if the Scenario has been locked for
              use.
            - ``cre_user`` and ``cre_date``—database user that created the
              Scenario, and creation time.
            - ``upd_user`` and ``upd_date``—user and time for last modification
              of the Scenario.
            - ``lock_user`` and ``lock_date``—user that locked the Scenario and
              lock time.
            - ``annotation``: description of the Scenario or changelog.
        """
        return pd.DataFrame(
            self._backend.get_scenarios(default, model, scen),
            columns=FIELDS["get_scenarios"],
        )

    def export_timeseries_data(
        self,
        path,
        default=True,
        model=None,
        scenario=None,
        variable=None,
        unit=None,
        region=None,
        export_all_runs=False,
    ):
        """Export timeseries data to CSV file across multiple scenarios.

        Refer :meth:`.add_timeseries` of :class:`Timeseries` to get more
        information about adding timeseries data to scenario.

        Parameters
        ----------
        path : os.PathLike
            File name to export data to; must have the suffix '.csv'.

            Result file will contain the following columns:

            - model
            - scenario
            - version
            - variable
            - unit
            - region
            - meta
            - subannual
            - year
            - value
        default : bool, optional
            :obj:`True` to include only TimeSeries versions marked as default.
        model: str, optional
            Only return data for this model name.
        scenario: str, optional
            Only return data for this scenario name.
        variable: list of str, optional
            Only return data for variable name(s) in this list.
        unit: list of str, optional
            Only return data for unit name(s) in this list.
        region: list of str, optional
            Only return data for region(s) in this list.
        export_all_runs: boolean, optional
            Export all existing model+scenario run combinations.


        Returns
        -------
        None
        """
        if export_all_runs and (model or scenario):
            raise ValueError(
                "Invalid arguments: "
                "export_all_runs cannot be used when providing "
                "a model or scenario."
            )
        filters = {
            "model": as_str_list(model) or [],
            "scenario": as_str_list(scenario) or [],
            "variable": as_str_list(variable) or [],
            "unit": as_str_list(unit) or [],
            "region": as_str_list(region) or [],
            "default": default,
            "export_all_runs": export_all_runs,
        }

        self._backend.write_file(path, ItemType.TS, filters=filters)

    def add_unit(self, unit, comment="None"):
        """Define a unit.

        Parameters
        ----------
        unit : str
            Name of the unit.
        comment : str, optional
            Annotation describing the unit or why it was added. The current
            database user and timestamp are appended automatically.
        """
        if unit in self.units():
            log.info(f"unit `{unit}` is already defined in the platform instance")
            return

        self._backend.set_unit(unit, comment)

    def units(self):
        """Return all units defined on the Platform.

        Returns
        -------
        numpy.ndarray of str
        """
        return self._backend.get_units()

    def regions(self):
        """Return all regions defined for the IAMC-style timeseries format
        including known synonyms.

        Returns
        -------
        :class:`pandas.DataFrame`
        """
        return pd.DataFrame(self._backend.get_nodes(), columns=FIELDS["get_nodes"])

    def _existing_node(self, name):
        """Check whether the node *name* has been defined.

        If :obj:`True`, log a warning and return True. Otherwise, return False.
        """
        for _, r in self.regions().iterrows():
            if r.region != name:
                continue

            log.warning(
                f"region {repr(name)} is already defined on the Platform"
                + (f" as a synonym for {repr(r.mapped_to)}" if r.mapped_to else "")
                + (f" under parent {repr(r.parent)}" if r.parent else "")
            )
            return True

        return False

    def add_region(self, region, hierarchy, parent="World"):
        """Define a region including a hierarchy level and a 'parent' region.

        .. tip::
           On a :class:`Platform` backed by a shared database, a region may
           already exist with a different spelling. Use :meth:`regions` first
           to check, and consider calling :meth:`add_region_synonym` instead.

        Parameters
        ----------
        region : str
            Name of the region.
        parent : str, default 'World'
            Assign a 'parent' region.
        hierarchy : str
            Hierarchy level of the region (e.g., country, R11, basin)
        """
        if not self._existing_node(region):
            self._backend.set_node(region, parent, hierarchy)

    def add_region_synonym(self, region, mapped_to):
        """Define a synonym for a `region`.

        When adding timeseries data using the synonym in the region column, it
        will be converted to `mapped_to`.

        Parameters
        ----------
        region : str
            Name of the region synonym.
        mapped_to : str
            Name of the region to which the synonym should be mapped.
        """
        if not self._existing_node(region):
            self._backend.set_node(region, synonym=mapped_to)

    def timeslices(self):
        """Return all subannual timeslices defined in this Platform instance.

        Timeslices are a way to represent subannual temporal resolution in timeseries
        data. A timeslice consists of a **name** (e.g., 'january', 'summer'), a
        **category** (e.g., 'months', 'seasons'), and a **duration** given relative to a
        full year.

        The category and duration do not have any functional relevance within the ixmp
        framework, but they may be useful for pre- or post-processing.  For example,
        they can be used to filter all timeslices of a certain category (e.g., all
        months) from the :class:`pandas.DataFrame` returned by this function or to
        aggregate subannual data to full-year results.

        A timeslice is related to the index set 'time' in a :class:`message_ix.Scenario`
        to indicate a subannual temporal dimension. Alas, timeslices and set elements of
        time have to be initialized/defined independently.

        See :meth:`add_timeslice` to initialize additional timeslices in the Platform
        instance.

        Returns
        -------
        :class:`pandas.DataFrame`
            DataFrame of timeslices, categories and duration
        """
        return pd.DataFrame(
            self._backend.get_timeslices(), columns=FIELDS["get_timeslices"]
        )

    def add_timeslice(self, name, category, duration):
        """Define a subannual timeslice including a category and duration.

        See :meth:`timeslices` for a detailed description of timeslices.

        Parameters
        ----------
        name : str
            Unique name of the timeslice.
        category : str
            Timeslice category (e.g. 'common', 'month', etc).
        duration : float
            Duration of timeslice as fraction of year.
        """
        slices = self.timeslices().set_index("name")
        if name in slices.index:
            msg = "timeslice `{}` already defined with duration {}"
            existing_duration = slices.loc[name].duration
            if not np.isclose(duration, existing_duration):
                raise ValueError(msg.format(name, existing_duration))
            log.info(msg.format(name, duration))
        else:
            self._backend.set_timeslice(name, category, duration)

    def check_access(self, user, models, access="view"):
        """Check access to specific models.

        Parameters
        ----------
        user: str
            Registered user name
        models : str or list of str
            Model(s) name
        access : str, optional
            Access type - view or edit

        Returns
        -------
        bool or dict of bool
        """
        models_list = as_str_list(models)
        result = self._backend.get_auth(user, models_list, access)
        if isinstance(models, str):
            return result[models]
        else:
            return {model: result.get(model) == 1 for model in models_list}


class TimeSeries:
    """Collection of data in time series format.

    TimeSeries is the parent/super-class of :class:`Scenario`.

    Parameters
    ----------
    mp : :class:`Platform`
        ixmp instance in which to store data.
    model : str
        Model name.
    scenario : str
        Scenario name.
    version : int or str, optional
        If omitted and a default version of the (`model`, `scenario`) has been
        designated (see :meth:`set_as_default`), load that version.
        If :class:`int`, load a specific version.
        If ``'new'``, create a new TimeSeries.
    annotation : str, optional
        A short annotation/comment used when ``version='new'``.
    """

    #: Name of the model associated with the TimeSeries
    model = None

    #: Name of the scenario associated with the TimeSeries
    scenario = None

    #: Version of the TimeSeries. Immutable for a specific instance.
    version = None

    def __init__(self, mp, model, scenario, version=None, annotation=None, **kwargs):
        # Check arguments
        if not isinstance(mp, Platform):
            raise TypeError("mp is not a valid `ixmp.Platform` instance")
        elif version and not (version == "new" or isinstance(version, int)):
            raise ValueError(f"version={repr(version)}")
        elif version == "new" and annotation is None:
            log.info(
                f"Missing annotation for new {self.__class__.__name__}"
                f" {model}/{scenario}"
            )
            annotation = ""

        # scheme= keyword argument only passed from Scenario.__init__;
        # otherwise must be None
        scheme = kwargs.get("scheme", None)
        if scheme:
            if self.__class__ is TimeSeries:
                raise TypeError("'scheme' argument to TimeSeries()")
            else:
                self.scheme = scheme

        # Set attributes
        self.model = model
        self.scenario = scenario

        # Store a weak reference to the Platform object. This reference is not
        # enough to keep the Platform alive, i.e. 'del mp' will work even while
        # this TimeSeries object lives.
        self.platform = mp if isinstance(mp, ProxyType) else proxy(mp)

        if version == "new":
            # Initialize a new object
            self._backend("init", annotation)
        else:
            # Retrieve an existing object
            self.version = version
            self._backend("get")

    def _backend(self, method, *args, **kwargs):
        """Convenience for calling *method* on the backend.

        The weak reference to the Platform object is used, if the Platform is
        still alive.
        """
        return self.platform._backend(self, method, *args, **kwargs)

    def __del__(self):
        # Instruct the back end to free memory associated with the TimeSeries
        try:
            self._backend("del_ts")
        except (AttributeError, ReferenceError):
            pass  # The Platform has already been garbage-collected

    # Transactions and versioning
    # functions for platform management
    def has_solution(self):
        # only Scenario class can have a solution
        return False

    def check_out(self, timeseries_only=False):
        """Check out the TimeSeries.

        Data in the TimeSeries can only be modified when it is in a checked-out
        state.

        See Also
        --------
        utils.maybe_check_out
        """
        self._backend("check_out", timeseries_only)

    def commit(self, comment):
        """Commit all changed data to the database.

        If the TimeSeries was newly created (with ``version='new'``),
        :attr:`version` is updated with a new version number assigned by the
        backend. Otherwise, :meth:`commit` does not change the :attr:`version`.

        Parameters
        ----------
        comment : str
            Description of the changes being committed.

        See Also
        --------
        utils.maybe_commit
        """
        self._backend("commit", comment)

    def discard_changes(self):
        """Discard all changes and reload from the database."""
        self._backend("discard_changes")

    def set_as_default(self):
        """Set the current :attr:`version` as the default."""
        self._backend("set_as_default")

    def is_default(self):
        """Return :obj:`True` if the :attr:`version` is the default version."""
        return self._backend("is_default")

    def last_update(self):
        """Get the timestamp of the last update/edit of this TimeSeries."""
        return self._backend("last_update")

    def run_id(self):
        """Get the run id of this TimeSeries."""
        return self._backend("run_id")

    # functions for importing and retrieving timeseries data

    def preload_timeseries(self):
        """Preload timeseries data to in-memory cache. Useful for bulk updates."""
        self._backend("preload")

    def add_timeseries(self, df, meta=False, year_lim=(None, None)):
        """Add data to the TimeSeries.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Data to add. `df` must have the following columns:

            - `region` or `node`
            - `variable`
            - `unit`

            Additional column names may be either of:

            - `year` and `value`—long, or 'tabular', format.
            - one or more specific years—wide, or 'IAMC' format.

            To support subannual temporal resolution of timeseries data, a
            column `subannual` is optional in `df`. The entries in this column
            must have been defined in the Platform instance using
            :meth:`add_timeslice` beforehand. If no column `subannual` is
            included in `df`, the data is assumed to contain yearly values.
            See :meth:`timeslices` for a detailed description of the feature.

        meta : bool, optional
            If :obj:`True`, store `df` as metadata. Metadata is treated
            specially when :meth:`Scenario.clone` is called for Scenarios
            created with ``scheme='MESSAGE'``.

        year_lim : tuple of (int or None, int or None`), optional
            Respectively, minimum and maximum years to add from *df*; data for
            other years is ignored.
        """
        meta = bool(meta)

        # Ensure consistent column names
        df = to_iamc_layout(df)

        if "value" in df.columns:
            # Long format; pivot to wide
            all_cols = [i for i in df.columns if i not in ["year", "value"]]
            df = pd.pivot_table(
                df, values="value", index=all_cols, columns=["year"]
            ).reset_index()
        df.set_index(["region", "variable", "unit", "subannual"], inplace=True)

        # Discard non-numeric columns, e.g. 'model', 'scenario', write warning about
        # non-expected cols to log
        year_cols = year_list(df.columns)
        other_cols = [
            i for i in df.columns if i not in ["model", "scenario"] + year_cols
        ]
        if len(other_cols) > 0:
            log.warning(f"Dropped extra column(s) {other_cols} from data")

        df = df.loc[:, year_cols]

        # Columns (year) as integer
        df.columns = df.columns.astype(int)

        # Identify columns to drop
        to_drop = set()
        if year_lim[0]:
            to_drop |= set(filter(lambda y: y < year_lim[0], df.columns))
        if year_lim[1]:
            to_drop |= set(filter(lambda y: y > year_lim[1], df.columns))

        df.drop(to_drop, axis=1, inplace=True)

        # Add one time series per row
        for (r, v, u, sa), data in df.iterrows():
            # Values as float; exclude NA
            self._backend("set_data", r, v, data.astype(float).dropna(), u, sa, meta)

    def timeseries(
        self,
        region=None,
        variable=None,
        unit=None,
        year=None,
        iamc=False,
        subannual="auto",
    ):
        """Retrieve timeseries data.

        Parameters
        ----------
        iamc : bool, optional
            Return data in wide/'IAMC' format. If :obj:`False`, return data in
            long/'tabular' format; see :meth:`add_timeseries`.
        region : str or list of str, optional
            Regions to include in returned data.
        variable : str or list of str, optional
            Variables to include in returned data.
        unit : str or list of str, optional
            Units to include in returned data.
        year : str or int or list of (str or int), optional
            Years to include in returned data.
        subannual : bool or 'auto', optional
            Whether to include column for sub-annual specification (if
            :class:`bool`); if 'auto', include column if sub-annual data (other
            than 'Year') exists in returned dataframe.

        Raises
        ------
        ValueError
            If `subannual` is :obj:`False` but Scenario has (filtered)
            sub-annual data.

        Returns
        -------
        pandas.DataFrame
            Specified data.
        """
        # Retrieve data, convert to pandas.DataFrame
        df = pd.DataFrame(
            self._backend(
                "get_data",
                as_str_list(region) or [],
                as_str_list(variable) or [],
                as_str_list(unit) or [],
                as_str_list(year) or [],
            ),
            columns=FIELDS["ts_get"],
        )
        df["model"] = self.model
        df["scenario"] = self.scenario

        # drop `subannual` column if not requested (False) or required ('auto')
        if subannual is not True:
            has_subannual = not all(df["subannual"] == "Year")
            if subannual is False and has_subannual:
                msg = (
                    "timeseries data has subannual values, ",
                    "use `subannual=True or 'auto'`",
                )
                raise ValueError(msg)
            if not has_subannual:
                df.drop("subannual", axis=1, inplace=True)

        if iamc:
            # Convert to wide format
            index = IAMC_IDX
            if "subannual" in df.columns:
                index = index + ["subannual"]
            df = df.pivot_table(index=index, columns="year")["value"].reset_index()
            df.columns.names = [None]

        return df

    def remove_timeseries(self, df):
        """Remove timeseries data from the TimeSeries instance.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Data to remove. `df` must have the following columns:

            - `region` or `node`
            - `variable`
            - `unit`
            - `year`
        """
        # Ensure consistent column names
        df = to_iamc_layout(df)

        id_cols = ["region", "variable", "unit", "subannual"]
        if "year" not in df.columns:
            # Reshape from wide to long format
            df = pd.melt(df, id_vars=id_cols, var_name="year", value_name="value")

        # Remove all years for a given (r, v, u) combination at once
        for (r, v, u, t), data in df.groupby(id_cols):
            self._backend("delete", r, v, t, data["year"].tolist(), u)

    def add_geodata(self, df):
        """Add geodata (layers) to the TimeSeries.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Data to add. `df` must have the following columns:

            - `region`
            - `variable`
            - `subannual`
            - `unit`
            - `year`
            - `value`
            - `meta`
        """
        for _, row in df.astype({"year": int, "meta": int}).iterrows():
            self._backend(
                "set_geo",
                row.region,
                row.variable,
                row.subannual,
                row.year,
                row.value,
                row.unit,
                row.meta,
            )

    def remove_geodata(self, df):
        """Remove geodata from the TimeSeries instance.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Data to remove. `df` must have the following columns:

            - `region`
            - `variable`
            - `unit`
            - `subannual`
            - `year`
        """
        # Remove all years for a given (r, v, t, u) combination at once
        for (r, v, t, u), data in df.groupby(
            ["region", "variable", "subannual", "unit"]
        ):
            self._backend("delete_geo", r, v, t, data["year"].tolist(), u)

    def get_geodata(self):
        """Fetch geodata and return it as dataframe.

        Returns
        -------
        :class:`pandas.DataFrame`
            Specified data.
        """
        # TODO remove astype here; this is the responsibility of Backend
        return (
            pd.DataFrame(self._backend("get_geo"), columns=FIELDS["ts_get_geo"])
            .reset_index(drop=True)
            .astype({"meta": "int64", "year": "int64"})
        )

    def read_file(self, path, firstyear=None, lastyear=None):
        """Read time series data from a CSV or Microsoft Excel file.

        Parameters
        ----------
        path : os.PathLike
            File to read. Must have suffix '.csv' or '.xlsx'.
        firstyear : int, optional
            Only read data from years equal to or later than this year.
        lastyear : int, optional
            Only read data from years equal to or earlier than this year.

        See also
        --------
        .Scenario.read_excel
        """
        self.platform._backend.read_file(
            Path(path),
            ItemType.TS,
            filters=dict(scenario=self),
            firstyear=firstyear,
            lastyear=lastyear,
        )


class Scenario(TimeSeries):
    """Collection of model-related data.

    See :class:`TimeSeries` for the meaning of parameters `mp`, `model`,
    `scenario`, `version`, and `annotation`.

    Parameters
    ----------
    scheme : str, optional
        Use an explicit scheme to initialize the new scenario. The
        :meth:`~.base.Model.initialize` method of the corresponding
        :class:`.Model` subclass in :data:`.MODELS` is used to initialize items
        in the Scenario.
    """

    #: Scheme of the Scenario.
    scheme = None

    def __init__(
        self,
        mp: Platform,
        model: str,
        scenario: str,
        version: Optional[Union[int, str]] = None,
        scheme: Optional[str] = None,
        annotation: Optional[str] = None,
        **model_init_args,
    ):
        # Check arguments
        if version == "new" and scheme is None:
            log.info(f"No scheme for new Scenario {model}/{scenario}")
            scheme = ""

        if "cache" in model_init_args:
            warn(
                "Scenario(..., cache=...) is deprecated; use Platform(..., "
                "cache=...) instead",
                DeprecationWarning,
            )
            model_init_args.pop("cache")

        # Call the parent constructor
        super().__init__(
            mp=mp,
            model=model,
            scenario=scenario,
            version=version,
            scheme=scheme,
            annotation=annotation,
        )

        if self.scheme == "MESSAGE" and self.__class__ is Scenario:
            # Loaded scenario has an improper scheme
            raise RuntimeError(
                f"{model}/{scenario} is a MESSAGE-scheme scenario; use "
                "message_ix.Scenario()"
            )

        # Retrieve the Model class correlating to the *scheme*
        model_class = get_model(self.scheme).__class__

        # Use the model class to initialize the Scenario
        model_class.initialize(self, **model_init_args)

    @classmethod
    def from_url(cls, url, errors="warn"):
        """Instantiate a Scenario given an ixmp-scheme URL.

        The following are equivalent::

            from ixmp import Platform, Scenario
            mp = Platform(name='example')
            scen = Scenario(mp 'model', 'scenario', version=42)

        and::

            from ixmp import Scenario
            scen, mp = Scenario.from_url('ixmp://example/model/scenario#42')

        Parameters
        ----------
        url : str
            See :meth:`parse_url <ixmp.utils.parse_url>`.
        errors : 'warn' or 'raise'
            If 'warn', a failure to load the Scenario is logged as a warning,
            and the platform is still returned. If 'raise', the exception
            is raised.

        Returns
        -------
        scenario, platform : 2-tuple of (Scenario, :class:`Platform`)
            The Scenario and Platform referred to by the URL.
        """
        assert errors in ("warn", "raise"), "errors= must be 'warn' or 'raise'"

        platform_info, scenario_info = parse_url(url)
        platform = Platform(**platform_info)

        try:
            scenario = cls(platform, **scenario_info)
        except Exception as e:
            if errors == "warn":
                log.warning(
                    f"{e.__class__.__name__}: {e.args[0]}\n"
                    f"when loading Scenario from url: {repr(url)}"
                )
                return None, platform
            else:
                raise
        else:
            return scenario, platform

    def check_out(self, timeseries_only=False):
        """Check out the Scenario.

        Raises
        ------
        ValueError
            If :meth:`has_solution` is :obj:`True`.

        See Also
        --------
        TimeSeries.check_out
        utils.maybe_check_out
        """
        if not timeseries_only and self.has_solution():
            raise ValueError(
                "This Scenario has a solution, "
                "use `Scenario.remove_solution()` or "
                "`Scenario.clone(..., keep_solution=False)`"
            )
        super().check_out(timeseries_only)

    def load_scenario_data(self):
        """Load all Scenario data into memory.

        Raises
        ------
        ValueError
            If the Scenario was instantiated with ``cache=False``.
        """
        if not getattr(self.platform._backend, "cache_enabled", False):
            raise ValueError("Cache must be enabled to load scenario data")

        for ix_type in "equ", "par", "set", "var":
            log.debug(f"Cache {repr(ix_type)} data")
            get_func = getattr(self, ix_type)
            for name in getattr(self, "{}_list".format(ix_type))():
                get_func(name)

    def idx_sets(self, name):
        """Return the list of index sets for an item (set, par, var, equ).

        Parameters
        ----------
        name : str
            name of the item
        """
        return self._backend("item_index", name, "sets")

    def idx_names(self, name):
        """Return the list of index names for an item (set, par, var, equ).

        Parameters
        ----------
        name : str
            name of the item
        """
        return self._backend("item_index", name, "names")

    def _keys(self, name, key_or_keys):
        if isinstance(key_or_keys, (list, pd.Series)):
            return as_str_list(key_or_keys)
        elif isinstance(key_or_keys, (pd.DataFrame, dict)):
            if isinstance(key_or_keys, dict):
                key_or_keys = pd.DataFrame.from_dict(key_or_keys, orient="columns")
            idx_names = self.idx_names(name)
            return [as_str_list(row, idx_names) for _, row in key_or_keys.iterrows()]
        else:
            return [str(key_or_keys)]

    def set_list(self):
        """List all defined sets."""
        return self._backend("list_items", "set")

    def has_set(self, name):
        """Check whether the scenario has a set *name*."""
        return name in self.set_list()

    def init_set(self, name, idx_sets=None, idx_names=None):
        """Initialize a new set.

        Parameters
        ----------
        name : str
            Name of the set.
        idx_sets : sequence of str or str, optional
            Names of other sets that index this set.
        idx_names : sequence of str or str, optional
            Names of the dimensions indexed by `idx_sets`.

        Raises
        ------
        ValueError
            If the set (or another object with the same *name*) already exists.
        RuntimeError
            If the Scenario is not checked out (see
            :meth:`~TimeSeries.check_out`).
        """
        idx_sets = as_str_list(idx_sets) or []
        idx_names = as_str_list(idx_names)
        return self._backend("init_item", "set", name, idx_sets, idx_names)

    def set(self, name, filters=None, **kwargs):
        """Return the (filtered) elements of a set.

        Parameters
        ----------
        name : str
            Name of the set.
        filters : dict
            Mapping of `dimension_name` → `elements`, where `dimension_name`
            is one of the `idx_names` given when the set was initialized (see
            :meth:`init_set`), and `elements` is an iterable of labels to
            include in the return value.

        Returns
        -------
        :class:`pandas.DataFrame`
        """
        return self._backend("item_get_elements", "set", name, filters)

    def add_set(self, name, key, comment=None):
        """Add elements to an existing set.

        Parameters
        ----------
        name : str
            Name of the set.
        key : str or iterable of str or dict or :class:`pandas.DataFrame`
            Element(s) to be added. If *name* exists, the elements are
            appended to existing elements.
        comment : str or iterable of str, optional
            Comment describing the element(s). If given, there must be the
            same number of comments as elements.

        Raises
        ------
        KeyError
            If the set *name* does not exist. :meth:`init_set` must be called
            before :meth:`add_set`.
        ValueError
            For invalid forms or combinations of *key* and *comment*.
        """
        # TODO expand docstring (here or in doc/source/api.rst) with examples,
        #      per test_core.test_add_set.

        if isinstance(key, list) and len(key) == 0:
            return  # No elements to add

        # Get index names for set *name*, may raise KeyError
        idx_names = self.idx_names(name)

        # Check arguments and convert to two lists: keys and comments
        if len(idx_names) == 0:
            # Basic set. Keys must be strings.
            if isinstance(key, (dict, pd.DataFrame)):
                raise ValueError(
                    "dict, DataFrame keys invalid for basic set {repr(name)}"
                )

            # Ensure keys is a list of str
            keys = as_str_list(key)
        else:
            # Set defined over 1+ other sets

            # Check for ambiguous arguments
            if comment and isinstance(key, (dict, pd.DataFrame)) and "comment" in key:
                raise ValueError("ambiguous; both key['comment'] and comment " "given")

            if isinstance(key, pd.DataFrame):
                # DataFrame of key values and perhaps comments
                try:
                    # Pop a 'comment' column off the DataFrame, convert to list
                    comment = key.pop("comment").to_list()
                except KeyError:
                    pass

                # Convert key to list of list of key values
                keys = []
                for row in key.to_dict(orient="records"):
                    keys.append(as_str_list(row, idx_names=idx_names))
            elif isinstance(key, dict):
                # Dict of lists of key values

                # Pop a 'comment' list from the dict
                comment = key.pop("comment", None)

                # Convert to list of list of key values
                keys = list(map(as_str_list, zip(*[key[i] for i in idx_names])))
            elif isinstance(key[0], str):
                # List of key values; wrap
                keys = [as_str_list(key)]
            elif isinstance(key[0], list):
                # List of lists of key values; convert to list of list of str
                keys = list(map(as_str_list, key))
            elif isinstance(key, str) and len(idx_names) == 1:
                # Bare key given for a 1D set; wrap for convenience
                keys = [[key]]
            else:
                # Other, invalid value
                raise ValueError(key)

        # Process comments to a list of str, or let them all be None
        comments = as_str_list(comment) if comment else repeat(None, len(keys))

        # Combine iterators to tuples. If the lengths are mismatched, the
        # sentinel value 'False' is filled in
        to_add = list(zip_longest(keys, comments, fillvalue=False))

        # Check processed arguments
        for e, c in to_add:
            # Check for sentinel values
            if e is False:
                raise ValueError(f"Comment {repr(c)} without matching key")
            elif c is False:
                raise ValueError(f"Key {repr(e)} without matching comment")
            elif len(idx_names) and len(idx_names) != len(e):
                raise ValueError(
                    f"{len(e)}-D key {repr(e)} invalid for "
                    f"{len(idx_names)}-D set {name}{repr(idx_names)}"
                )

        # Send to backend
        elements = ((kc[0], None, None, kc[1]) for kc in to_add)
        self._backend("item_set_elements", "set", name, elements)

    def remove_set(self, name, key=None):
        """Delete set elements or an entire set.

        Parameters
        ----------
        name : str
            Name of the set to remove (if `key` is :obj:`None`) or from which
            to remove elements.
        key : :class:`pandas.DataFrame` or list of str, optional
            Elements to be removed from set `name`.
        """
        if key is None:
            self._backend("delete_item", "set", name)
        else:
            self._backend("item_delete_elements", "set", name, self._keys(name, key))

    def par_list(self):
        """List all defined parameters."""
        return self._backend("list_items", "par")

    def has_par(self, name):
        """Check whether the scenario has a parameter with that name."""
        return name in self.par_list()

    def init_par(self, name, idx_sets, idx_names=None):
        """Initialize a new parameter.

        Parameters
        ----------
        name : str
            Name of the parameter.
        idx_sets : sequence of str or str, optional
            Names of sets that index this parameter.
        idx_names : sequence of str or str, optional
            Names of the dimensions indexed by `idx_sets`.
        """
        idx_sets = as_str_list(idx_sets) or []
        idx_names = as_str_list(idx_names)
        return self._backend("init_item", "par", name, idx_sets, idx_names)

    def par(self, name, filters=None, **kwargs):
        """Return parameter data.

        If *filters* is provided, only a subset of data, matching the filters,
        is returned.

        Parameters
        ----------
        name : str
            Name of the parameter
        filters : dict (str -> list of str), optional
            Index names mapped to lists of index set elements. Elements not
            appearing in the respective index set(s) are silently ignored.
        """
        if len(kwargs):
            raise DeprecationWarning(
                "ignored kwargs to Scenario.par(); will raise TypeError in 4.0"
            )
        return self._backend("item_get_elements", "par", name, filters)

    def items(self, type=ItemType.PAR, filters=None):
        """Iterate over model data items.

        Parameters
        ----------
        type : ItemType, optional
            Types of items to iterate, e.g. :data:`ItemType.PAR` for
            parameters, the only value currently supported.
        filters : dict, optional
            Filters for values along dimensions; same as the `filters` argument
            to :meth:`par`.

        Yields
        ------
        (str, object)
            Tuples of item name and data.
        """
        if type != ItemType.PAR:
            raise NotImplementedError(
                f"Scenario.items(type={type}); only ItemType.PAR is supported"
            )

        filters = filters or dict()

        names = sorted(self.par_list())

        for name in sorted(names):
            idx_names = set(self.idx_names(name))
            if len(filters) and not set(filters.keys()) & idx_names:
                # No overlap between the filters and this item's dimensions
                continue

            # Retrieve the data, reducing the filters to only the dimensions of
            # the item
            yield name, self.par(
                name, filters={k: v for k, v in filters.items() if k in idx_names}
            )

    def add_par(
        self,
        name: str,
        key_or_data=None,
        value=None,
        unit: str = None,
        comment: str = None,
    ):
        """Set the values of a parameter.

        Parameters
        ----------
        name : str
            Name of the parameter.
        key_or_data : str or iterable of str or range or dict or
                      :class:`pandas.DataFrame`
            Element(s) to be added.
        value : numeric or iterable of numeric, optional
            Values.
        unit : str or iterable of str, optional
            Unit symbols.
        comment : str or iterable of str, optional
            Comment(s) for the added values.
        """
        # Number of dimensions in the index of *name*
        idx_names = self.idx_names(name)
        N_dim = len(idx_names)

        # Convert valid forms of arguments to pd.DataFrame
        if isinstance(key_or_data, dict):
            # dict containing data
            data = pd.DataFrame.from_dict(key_or_data, orient="columns")
        elif isinstance(key_or_data, pd.DataFrame):
            data = key_or_data.copy()
            if value is not None:
                if "value" in data.columns:
                    raise ValueError("both key_or_data.value and value supplied")
                else:
                    data["value"] = value
        else:
            # One or more keys; convert to a list of strings
            if isinstance(key_or_data, range):
                key_or_data = list(key_or_data)
            keys = self._keys(name, key_or_data)

            # Check the type of value
            if isinstance(value, (float, int)):
                # Single value

                if N_dim > 1 and len(keys) == N_dim:
                    # Ambiguous case: ._key() above returns ['dim_0', 'dim_1'],
                    # when we really want [['dim_0', 'dim_1']]
                    keys = [keys]

                # Use the same value for all keys
                values = [float(value)] * len(keys)
            else:
                # Multiple values
                values = value

            data = pd.DataFrame(zip(keys, values), columns=["key", "value"])
            if data.isna().any(axis=None):
                raise ValueError("Length mismatch between keys and values")

        # Column types
        types = {
            "key": str if N_dim == 1 else object,
            "value": float,
            "unit": str,
            "comment": str,
        }

        # Further handle each column
        if "key" not in data.columns:
            # Form the 'key' column from other columns
            if N_dim > 1 and len(data):
                data["key"] = data.apply(
                    partial(as_str_list, idx_names=idx_names), axis=1
                )
            else:
                data["key"] = data[idx_names[0]]

        if "value" not in data.columns:
            raise ValueError("no parameter values supplied")

        if "unit" not in data.columns:
            # Broadcast single unit across all values. pandas raises ValueError
            # if *unit* is iterable but the wrong length
            data["unit"] = unit or "???"

        if "comment" not in data.columns:
            if comment:
                # Broadcast single comment across all values. pandas raises
                # ValueError if *comment* is iterable but the wrong length
                data["comment"] = comment
            else:
                # Store a 'None' comment
                data["comment"] = None
                types.pop("comment")

        # Convert types, generate tuples
        elements = map(
            lambda e: (e.key, e.value, e.unit, e.comment),
            data.astype(types).itertuples(),
        )

        # Store
        self._backend("item_set_elements", "par", name, elements)

    def init_scalar(self, name, val, unit, comment=None):
        """Initialize a new scalar.

        Parameters
        ----------
        name : str
            Name of the scalar
        val : number
            Initial value of the scalar.
        unit : str
            Unit of the scalar.
        comment : str, optional
            Description of the scalar.
        """
        self.init_par(name, [], [])
        self.change_scalar(name, val, unit, comment)

    def scalar(self, name):
        """Return the value and unit of a scalar.

        Parameters
        ----------
        name : str
            Name of the scalar.

        Returns
        -------
        {'value': value, 'unit': unit}
        """
        return self._backend("item_get_elements", "par", name, None)

    def change_scalar(self, name, val, unit, comment=None):
        """Set the value and unit of a scalar.

        Parameters
        ----------
        name : str
            Name of the scalar.
        val : number
            New value of the scalar.
        unit : str
            New unit of the scalar.
        comment : str, optional
            Description of the change.
        """
        self._backend(
            "item_set_elements", "par", name, [(None, float(val), unit, comment)]
        )

    def remove_par(self, name, key=None):
        """Remove parameter values or an entire parameter.

        Parameters
        ----------
        name : str
            Name of the parameter.
        key : dataframe or key list or concatenated string, optional
            elements to be removed
        """
        if key is None:
            self._backend("delete_item", "par", name)
        else:
            self._backend("item_delete_elements", "par", name, self._keys(name, key))

    def var_list(self):
        """List all defined variables."""
        return self._backend("list_items", "var")

    def has_var(self, name):
        """Check whether the scenario has a variable with that name."""
        return name in self.var_list()

    def init_var(self, name, idx_sets=None, idx_names=None):
        """Initialize a new variable.

        Parameters
        ----------
        name : str
            Name of the variable.
        idx_sets : sequence of str or str, optional
            Name(s) of index sets for a 1+-dimensional variable.
        idx_names : sequence of str or str, optional
            Names of the dimensions indexed by `idx_sets`.
        """
        idx_sets = as_str_list(idx_sets) or []
        idx_names = as_str_list(idx_names)
        return self._backend("init_item", "var", name, idx_sets, idx_names)

    def var(self, name, filters=None, **kwargs):
        """Return a dataframe of (filtered) elements for a specific variable.

        Parameters
        ----------
        name : str
            name of the variable
        filters : dict
            index names mapped list of index set elements
        """
        return self._backend("item_get_elements", "var", name, filters)

    def equ_list(self):
        """List all defined equations."""
        return self._backend("list_items", "equ")

    def init_equ(self, name, idx_sets=None, idx_names=None):
        """Initialize a new equation.

        Parameters
        ----------
        name : str
            Name of the equation.
        idx_sets : sequence of str or str, optional
            Name(s) of index sets for a 1+-dimensional variable.
        idx_names : sequence of str or str, optional
            Names of the dimensions indexed by `idx_sets`.
        """
        idx_sets = as_str_list(idx_sets) or []
        idx_names = as_str_list(idx_names)
        return self._backend("init_item", "equ", name, idx_sets, idx_names)

    def has_equ(self, name):
        """Check whether the scenario has an equation with that name."""
        return name in self.equ_list()

    def equ(self, name, filters=None, **kwargs):
        """Return a dataframe of (filtered) elements for a specific equation.

        Parameters
        ----------
        name : str
            name of the equation
        filters : dict
            index names mapped list of index set elements
        """
        return self._backend("item_get_elements", "equ", name, filters)

    def clone(
        self,
        model=None,
        scenario=None,
        annotation=None,
        keep_solution=True,
        shift_first_model_year=None,
        platform=None,
    ):
        """Clone the current scenario and return the clone.

        If the (`model`, `scenario`) given already exist on the
        :class:`Platform`, the `version` for the cloned Scenario follows the
        last existing version. Otherwise, the `version` for the cloned Scenario
        is 1.

        .. note::
            :meth:`clone` does not set or alter default versions. This means
            that a clone to new (`model`, `scenario`) names has no default
            version, and will not be returned by
            :meth:`Platform.scenario_list` unless `default=False` is given.

        Parameters
        ----------
        model : str, optional
            New model name. If not given, use the existing model name.
        scenario : str, optional
            New scenario name. If not given, use the existing scenario name.
        annotation : str, optional
            Explanatory comment for the clone commit message to the database.
        keep_solution : bool, optional
            If :py:const:`True`, include all timeseries data and the solution
            (vars and equs) from the source scenario in the clone.
            If :py:const:`False`, only include timeseries data marked
            `meta=True` (see :meth:`TimeSeries.add_timeseries`).
        shift_first_model_year: int, optional
            If given, all timeseries data in the Scenario is omitted from the
            clone for years from `first_model_year` onwards. Timeseries data
            with the `meta` flag (see :meth:`TimeSeries.add_timeseries`) are
            cloned for all years.
        platform : :class:`Platform`, optional
            Platform to clone to (default: current platform)
        """
        if shift_first_model_year is not None:
            if keep_solution:
                log.warning("Override keep_solution=True for shift_first_model_year")
                keep_solution = False

        platform = platform or self.platform
        model = model or self.model
        scenario = scenario or self.scenario

        args = [platform, model, scenario, annotation, keep_solution]
        if check_year(shift_first_model_year, "first_model_year"):
            args.append(shift_first_model_year)

        return self._backend("clone", *args)

    def has_solution(self):
        """Return :obj:`True` if the Scenario has been solved.

        If ``has_solution() == True``, model solution data exists in the db.
        """
        return self._backend("has_solution")

    def remove_solution(self, first_model_year=None):
        """Remove the solution from the scenario

        This function removes the solution (variables and equations) and
        timeseries data marked as `meta=False` from the scenario
        (see :meth:`TimeSeries.add_timeseries`).

        Parameters
        ----------
        first_model_year: int, optional
            If given, timeseries data marked as `meta=False` is removed
            only for years from `first_model_year` onwards.

        Raises
        ------
        ValueError
            If Scenario has no solution or if `first_model_year` is not `int`.
        """
        if self.has_solution():
            check_year(first_model_year, "first_model_year")
            self._backend("clear_solution", first_model_year)
        else:
            raise ValueError("This Scenario does not have a solution!")

    def solve(self, model=None, callback=None, cb_kwargs={}, **model_options):
        """Solve the model and store output.

        ixmp 'solves' a model by invoking the run() method of a :class:`.Model`
        subclass—for instance, :meth:`.GAMSModel.run`. Depending on the
        underlying model code, different steps are taken; see each model class
        for details. In general:

        1. Data from the Scenario are written to a **model input file**.
        2. Code or an external program is invoked to perform calculations or
           optimizations, **solving the model**.
        3. Data representing the model outputs or solution are read from a
           **model output file** and stored in the Scenario.

        If the optional argument `callback` is given, then additional steps are
        performed:

        4. Execute the `callback` with the Scenario as an argument. The
           Scenario has an `iteration` attribute that stores the number of
           times the underlying model has been solved (#2).
        5. If the `callback` returns :obj:`False` or similar, iterate by
           repeating from step #1. Otherwise, exit.

        Parameters
        ----------
        model : str
            model (e.g., MESSAGE) or GAMS file name (excluding '.gms')
        callback : callable, optional
            Method to execute arbitrary non-model code. Must accept a single
            argument: the Scenario. Must return a non-:obj:`False` value to
            indicate convergence.
        cb_kwargs : dict, optional
            Keyword arguments to pass to `callback`.
        model_options :
            Keyword arguments specific to the `model`. See :class:`.GAMSModel`.

        Warns
        -----
        UserWarning
            If `callback` is given and returns :obj:`None`. This may indicate
            that the user has forgotten a ``return`` statement, in which case
            the iteration will continue indefinitely.

        Raises
        ------
        ValueError
            If the Scenario has already been solved.
        """
        if self.has_solution():
            raise ValueError(
                "This Scenario has already been solved, ",
                "use `remove_solution()` first!",
            )

        # Instantiate a model
        model = get_model(model or self.scheme, **model_options)

        # Validate *callback* argument
        if callback is not None and not callable(callback):
            raise ValueError(f"callback={repr(callback)} is not callable")
        elif callback is None:
            # Make the callback a no-op
            def callback(scenario, **kwargs):
                return True

        # Flag to warn if the *callback* appears not to return anything
        warn_none = True

        # Iterate until convergence
        while True:
            model.run(self)

            # Store an iteration number to help the callback
            if not hasattr(self, "iteration"):
                self.iteration = 0

            self.iteration += 1

            # Invoke the callback
            cb_result = callback(self, **cb_kwargs)

            if cb_result is None and warn_none:
                warn(
                    "solve(callback=...) argument returned None; will loop "
                    "indefinitely unless True is returned."
                )
                # Don't repeat the warning
                warn_none = False

            if cb_result:
                # Callback indicates convergence is reached
                break

    def get_meta(self, name=None):
        """Get scenario meta.

        Parameters
        ----------
        name : str, optional
            meta category name
        """
        all_meta = self.platform._backend.get_meta(
            self.model, self.scenario, self.version
        )
        return all_meta[name] if name else all_meta

    def set_meta(self, name_or_dict, value=None):
        """Set scenario meta.

        Parameters
        ----------
        name_or_dict : str or dict
            If the argument is dict, it used as a mapping of meta
            categories (names) to values. Otherwise, use the argument
            as the meta category name.
        value : str or number or bool, optional
            Meta category value.
        """
        if not isinstance(name_or_dict, dict):
            if isinstance(name_or_dict, str):
                name_or_dict = {name_or_dict: value}
            else:
                msg = (
                    "Unsupported parameter type of name_or_dict: %s. "
                    "Supported parameter types for name_or_dict are "
                    "String and Dictionary"
                ) % type(name_or_dict)
                raise ValueError(msg)
        self.platform._backend.set_meta(
            name_or_dict, self.model, self.scenario, self.version
        )

    def delete_meta(self, *args, **kwargs):
        """Remove scenario meta.

        .. deprecated:: 3.1

           Use :meth:`remove_meta()`.

        Parameters
        ----------
        name : str or list of str
            Either single meta key or list of keys.
        """
        warn("Scenario.delete_meta(); use remove_meta()", DeprecationWarning)
        self.remove_meta(*args, **kwargs)

    def remove_meta(self, name):
        """Remove scenario meta.

        Parameters
        ----------
        name : str or list of str
            Either single meta key or list of keys.
        """
        if isinstance(name, str):
            name = [name]
        self.platform._backend.remove_meta(
            name, self.model, self.scenario, self.version
        )

    # Input and output
    def to_excel(
        self, path, items=ItemType.SET | ItemType.PAR, filters=None, max_row=None
    ):
        """Write Scenario to a Microsoft Excel file.

        Parameters
        ----------
        path : os.PathLike
            File to write. Must have suffix :file:`.xlsx`.
        items : ItemType, optional
            Types of items to write. Either :attr:`.SET` | :attr:`.PAR` (i.e.
            only sets and parameters), or :attr:`.MODEL` (also variables and
            equations, i.e. model solution data).
        filters : dict, optional
            Filters for values along dimensions; same as the `filters` argument
            to :meth:`par`.
        max_row: int, optional
            Maximum number of rows in each sheet. If the number of elements in
            an item exceeds this number or :data:`.EXCEL_MAX_ROWS`, then an
            item is written to multiple sheets named, e.g. 'foo', 'foo(2)',
            'foo(3)', etc.

        See also
        --------
        :ref:`excel-data-format`
        read_excel
        """
        # Default filters: empty dict
        filters = filters or dict()

        # Select the current scenario
        filters["scenario"] = self

        # Invoke the backend method
        self.platform._backend.write_file(
            Path(path), items, filters=filters, max_row=max_row
        )

    def read_excel(self, path, add_units=False, init_items=False, commit_steps=False):
        """Read a Microsoft Excel file into the Scenario.

        Parameters
        ----------
        path : os.PathLike
            File to read. Must have suffix '.xlsx'.
        add_units : bool, optional
            Add missing units, if any, to the Platform instance.
        init_items : bool, optional
            Initialize sets and parameters that do not already exist in the
            Scenario.
        commit_steps : bool, optional
            Commit changes after every data addition.

        See also
        --------
        :ref:`excel-data-format`
        .TimeSeries.read_file
        to_excel
        """
        self.platform._backend.read_file(
            Path(path),
            ItemType.MODEL,
            filters=dict(scenario=self),
            add_units=add_units,
            init_items=init_items,
            commit_steps=commit_steps,
        )


def to_iamc_layout(df):
    """Transform *df* to a standard IAMC layout.

    The returned object has:

    - Any (Multi)Index levels reset as columns.
    - Lower-case column names 'region', 'variable', 'subannual', and 'unit'.
    - If not present in *df*, the value 'Year' in the 'subannual' column.

    Parameters
    ----------
    df : pandas.DataFrame
        May have a 'node' column, which will be renamed to 'region'.

    Returns
    -------
    pandas.DataFrame

    Raises
    ------
    ValueError
        If 'region', 'variable', or 'unit' is not among the column names.
    """
    # Reset the index if meaningful entries are included there
    if not list(df.index.names) == [None]:
        df.reset_index(inplace=True)

    # Rename columns in lower case, and transform 'node' to 'region'
    cols = {c: str(c).lower() for c in df.columns}
    cols.update(node="region")
    df = df.rename(columns=cols)

    required_cols = ["region", "variable", "unit"]
    missing = list(set(required_cols) - set(df.columns))
    if len(missing):
        raise ValueError(f"missing required columns {repr(missing)}")

    # Add a column 'subannual' with the default value
    if "subannual" not in df.columns:
        df["subannual"] = "Year"

    return df



"""Testing Juypter notebooks."""
import os
import sys

import pytest

nbformat = pytest.importorskip("nbformat")


def run_notebook(nb_path, tmp_path, env=None, kernel=None, allow_errors=False):
    """Execute a Jupyter notebook via :mod:`nbclient` and collect output.

    Parameters
    ----------
    nb_path : path-like
        The notebook file to execute.
    tmp_path : path-like
        A directory in which to create temporary output.
    env : dict-like, optional
        Execution environment for :mod:`nbclient`. Default: :obj:`os.environ`.
    kernel : str, optional
        Jupyter kernel to use. Default: 'python2' or 'python3', matching the current
        Python version.
    allow_errors : bool, optional
        Whether to pass the ``--allow-errors`` option to :mod:`nbclient`. If
        :obj:`True`, the execution always succeeds, and cell output contains exception
        information.

    Returns
    -------
    nb : :class:`nbformat.NotebookNode`
        Parsed and executed notebook.
    errors : list
        Any execution errors.
    """
    import nbformat
    from nbclient import NotebookClient

    # Workaround for https://github.com/jupyter/nbclient/issues/85
    if (
        sys.version_info[0] == 3
        and sys.version_info[1] >= 8
        and sys.platform.startswith("win")
    ):
        import asyncio

        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    # Read the notebook
    nb = nbformat.read(nb_path, as_version=4)

    # Create a client and use it to execute the notebook
    client = NotebookClient(
        nb,
        timeout=10,
        kernel_name=kernel or f"python{sys.version_info[0]}",
        allow_errors=allow_errors,
        resources=dict(metadata=dict(path=tmp_path)),
    )

    # Execute the notebook.
    # `env` is passed from nbclient to jupyter_client.launcher.launch_kernel()
    client.execute(env=env or os.environ.copy())

    # Retrieve error information from cells
    errors = [
        output
        for cell in nb.cells
        if "outputs" in cell
        for output in cell["outputs"]
        if output.output_type == "error"
    ]

    return nb, errors


def get_cell_output(nb, name_or_index, kind="data"):
    """Retrieve a cell from `nb` according to its metadata `name_or_index`:

    The Jupyter notebook format allows specifying a document-wide unique 'name' metadata
    attribute for each cell:

    https://nbformat.readthedocs.io/en/latest/format_description.html
    #cell-metadata

    Return the cell matching `name_or_index` if :class:`str`; or the cell at the
    :class:`int` index; or raise :class:`ValueError`.

    Parameters
    ----------
    kind : str, optional
        Kind of cell output to retrieve. For 'data', the data in format 'text/plain' is
        run through :func:`eval`. To retrieve an exception message, use 'evalue'.
    """
    if isinstance(name_or_index, int):
        cell = nb.cells[name_or_index]
    else:
        for i, _cell in enumerate(nb.cells):
            try:
                if _cell.metadata.jupyter.name == name_or_index:
                    cell = _cell
                    break
            except AttributeError:
                continue

    try:
        result = cell["outputs"][0][kind]
    except NameError:  # pragma: no cover
        raise ValueError(f"no cell named {name_or_index}")
    else:
        return eval(result["text/plain"]) if kind == "data" else result



"""Performance testing."""
import logging
from collections import namedtuple

try:
    import resource

    has_resource_module = True
except ImportError:
    # Windows
    has_resource_module = False

import numpy as np
import pytest

log = logging.getLogger(__name__)


#: Data structure for memory information used by :meth:`memory_usage`.
MemInfo = namedtuple("MemInfo", "profiled max_rss jvm_total jvm_free jvm_used python")


def format_meminfo(arr, cls=float):
    """Return a namedtuple for `arr`, with values as `cls`."""
    # Format strings
    cls = "{: >7.2f}".format if cls is str else cls
    return MemInfo(*map(cls, arr))


# Variables for memory_usage
_COUNT = 0
_PREV = np.zeros(6)
_RT = None


def memory_usage(message="", reset=False):
    """Profile memory usage from within a test function.

    The Python package ``memory_profiler`` and :mod:`jpype` are used to report memory
    usage. A message is logged at the ``DEBUG`` level, similar to::

       DEBUG    ixmp.testing:testing.py:527  42 <message>
       MemInfo(profiled=' 533.76', max_rss=' 534.19', jvm_total=' 213.50',
               jvm_free='  79.22', jvm_used=' 134.28', python='399.48')
       MemInfo(profiled='   0.14', max_rss='   0.00', jvm_total='   0.00',
               jvm_free=' -37.75', jvm_used='  37.75', python=' -37.61')

    Parameters
    ----------
    message : str, optional
        A string added to the log message, to identify points in profiled code.
    reset : bool, optional
        If :obj:`True`, start profiling anew.

    Returns
    -------
    collections.namedtuple
        A ``MemInfo`` tuple with the following fields, all in MiB:

        - ``profiled``: the instantaneous memory usage reported by memory_profiler.
        - ``max_rss``: the maximum resident set size (i.e. the maximum over the entire
          life of the process) reported by :meth:`resource.getrusage`.
        - ``jvm_total``: total memory allocated for the Java Virtual Machine (JVM)
          underlying JPype, used by :class:`.JDBCBackend`; the same as
          ``java.lang.Runtime.getRuntime().totalMemory()``.
        - ``jvm_free``: memory allocated to the JVM that is free.
        - ``jvm_used``: memory used by the JVM, i.e. ``jvm_total`` minus ``jvm_free``.
        - ``python``: a rough estimate of Python memory usage, i.e. ``profiled`` minus
          ``jvm_used``. This may not be accurate.
    """
    import memory_profiler

    from ixmp.backend.jdbc import java

    global _COUNT, _PREV, _RT

    if reset:
        _COUNT = 0
        _PREV = np.zeros(6)
    else:
        _COUNT += 1

    try:
        # Get the Java runtime
        runtime = _RT or java.Runtime.getRuntime()
    except AttributeError:
        # JVM not loaded
        runtime = None

    # Collect data
    result = [
        memory_profiler.memory_usage()[0],
        resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024,  # to MiB
    ]
    if runtime:
        _RT = runtime
        result.extend(
            [
                _RT.totalMemory() / 1024 ** 2,
                _RT.freeMemory() / 1024 ** 2,
            ]
        )
    else:
        result.extend([0, 0])

    # JVM total - JVM free = JVM used
    result.append(result[-2] - result[-1])

    # Python used - JVM used = only-Python used?
    result.append(result[0] - result[-1])

    # Convert to a numpy array
    result = np.array(result)

    # Difference versus previous call
    delta = result - _PREV

    # Store current *result* to compute *delta* in subsequent calls
    _PREV = result

    # Log the results
    msg = "\n".join(
        [
            f"{_COUNT:3d} {message}",
            repr(format_meminfo(result, str)),
            repr(format_meminfo(delta, str)),
        ]
    )
    log.debug(msg)

    # Return the current result
    return format_meminfo(result)


@pytest.fixture(scope="function")
def resource_limit(request):
    """A fixture that limits Python :mod:`resources`.

    See the documentation (``pytest --help``) for the ``--resource-limit`` command-line
    option that selects (1) the specific resource and (2) the level of the limit.

    The original limit, if any, is reset after the test function in which the fixture is
    used.
    """
    if not has_resource_module:
        pytest.skip("Python module 'resource' not available (non-Unix OS)")

    name, value = request.config.getoption("--resource-limit").split(":")
    res = getattr(resource, f"RLIMIT_{name.upper()}")
    value = int(value)

    if res in (resource.RLIMIT_AS, resource.RLIMIT_DATA, resource.RLIMIT_RSS):
        value = value * 1024 ** 2  # MiB → bytes

    if value > 0:
        # Store existing limit
        before = resource.getrlimit(res)

        log.debug(f"Change {res} from {before} to ({value}, {before[1]})")
        resource.setrlimit(res, (value, before[1]))

    try:
        yield
    finally:
        if value > 0:
            log.debug(f"Restore {res} to {before}")
            resource.setrlimit(res, before)



import logging
import re

import pytest

from ixmp import Scenario
from ixmp.model.base import Model, ModelError
from ixmp.model.dantzig import DantzigModel
from ixmp.model.gams import gams_version
from ixmp.testing import assert_logs, make_dantzig


def test_base_model():
    # An incomplete Backend subclass can't be instantiated
    class M1(Model):
        pass

    with pytest.raises(
        TypeError, match="Can't instantiate abstract class M1 " "with abstract methods"
    ):
        M1()


def test_model_initialize(test_mp, caplog):
    # Model.initialize runs on an empty Scenario
    s = make_dantzig(test_mp)
    b1 = s.par("b")
    assert len(b1) == 3

    # Modify a value for 'b'
    s.check_out()
    new_value = 301
    s.add_par("b", "chicago", new_value, "cases")
    s.commit("Overwrite b(chicago)")

    # Model.initialize runs on an already-initialized Scenario, without error
    DantzigModel.initialize(s, with_data=True)

    # Data has the same length...
    b2 = s.par("b")
    assert len(b2) == 3
    # ...but modified value(s) are not overwritten
    assert (b2.query("j == 'chicago'")["value"] == new_value).all()

    # Unrecognized Scenario(scheme=...) is initialized using the base method, a
    # no-op
    messages = [
        "No scheme for new Scenario model-name/scenario-name",
        "No initialization for None-scheme Scenario",
    ]
    with assert_logs(caplog, messages, at_level=logging.DEBUG):
        Scenario(test_mp, model="model-name", scenario="scenario-name", version="new")

    with assert_logs(
        caplog, "No initialization for 'foo'-scheme Scenario", at_level=logging.DEBUG
    ):
        Scenario(
            test_mp,
            model="model-name",
            scenario="scenario-name",
            version="new",
            scheme="foo",
        )

    # Keyword arguments to Scenario(...) that are not recognized by
    # Model.initialize() raise an intelligible exception
    with pytest.raises(TypeError, match="unexpected keyword argument 'bad_arg1'"):
        Scenario(
            test_mp,
            model="model-name",
            scenario="scenario-name",
            version="new",
            scheme="unknown",
            bad_arg1=111,
        )

    with pytest.raises(TypeError, match="unexpected keyword argument 'bad_arg2'"):
        Scenario(
            test_mp,
            model="model-name",
            scenario="scenario-name",
            version="new",
            scheme="dantzig",
            with_data=True,
            bad_arg2=222,
        )

    # Replace b[j] with a parameter of the same name, but different indices
    s.check_out()
    s.remove_par("b")
    s.init_par("b", idx_sets=["i"], idx_names=["i_dim"])

    # Logs an error message
    with assert_logs(caplog, "Existing index sets of 'b' ['i'] do not match ['j']"):
        DantzigModel.initialize(s)


def test_gams_version():
    # Returns a version string like X.Y.Z
    assert len(gams_version().split(".")) == 3


class TestGAMSModel:
    @pytest.fixture(scope="class")
    def dantzig(self, test_mp):
        yield make_dantzig(test_mp)

    @pytest.mark.parametrize("char", r'<>"/\|?*')
    def test_filename_invalid_char(self, dantzig, char):
        """Model can be solved with invalid character names."""
        name = f"foo{char}bar"
        s = dantzig.clone(model=name, scenario=name)

        # Indirectly test backend.write_file("….gdx")
        # This name_ keyword argument ends up received to GAMSModel.__init__ and sets
        # the GAMSModel.model_name attribute, and in turn the GDX file names used.
        s.solve(name_=name, quiet=True)

    @pytest.mark.parametrize(
        "kwargs",
        [
            dict(comment=None),
            dict(equ_list=None, var_list=["x"]),
            dict(equ_list=["demand", "supply"], var_list=[]),
        ],
        ids=["null-comment", "null-list", "empty-list"],
    )
    def test_GAMSModel_solve(test_data_path, dantzig, kwargs):
        """Options to GAMSModel are handled without error."""
        dantzig.clone().solve(**kwargs, quiet=True)

    def test_error_message(self, test_data_path, test_mp):
        """GAMSModel.solve() displays a user-friendly message on error."""
        # Empty Scenario
        s = Scenario(test_mp, model="foo", scenario="bar", version="new")
        s.commit("Initial commit")

        # Expected paths for error message
        paths = map(
            lambda name: re.escape(str(test_data_path.joinpath(name))),
            ["_abort.lst", "default_in.gdx"],
        )

        with pytest.raises(
            ModelError,
            match="""GAMS errored with return code 2:
    There was a compilation error

For details, see the terminal output above, plus:
Listing   : {}
Input data: {}""".format(
                *paths
            ),
        ):
            s.solve(model_file=test_data_path / "_abort.gms", use_temp_dir=False)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full list see
# the documentation: https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory, add
# these directories to sys.path here. If the directory is relative to the documentation
# root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# Import so that autodoc can find code
import ixmp
import ixmp.testing

# -- Project information ---------------------------------------------------------------

project = "ix modeling platform"
copyright = "2017–2021, IIASA Energy, Climate, and Environment (ECE) program"
author = "ixmp Developers"


# -- General configuration -------------------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinxcontrib.bibtex",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "README.rst", "Thumbs.db", ".DS_Store"]

# A string of reStructuredText that will be included at the beginning of every
# source file that is read.
version = ixmp.__version__
rst_prolog = r"""
.. |MESSAGEix| replace:: MESSAGE\ :emphasis:`ix`

.. |ixmp| replace:: :emphasis:`ix modeling platform`

.. |version| replace:: {}

.. role:: strike

.. role:: underline

""".format(
    version
)

# -- Options for HTML output -----------------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# A list of CSS files.
html_css_files = ["custom.css"]

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logo_white.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for sphinx.ext.extlinks ---------------------------------------------------

extlinks = {
    "issue": ("https://github.com/iiasa/ixmp/issue/%s", "#"),
    "pull": ("https://github.com/iiasa/ixmp/pull/%s", "PR #"),
}

# -- Options for sphinx.ext.intersphinx ------------------------------------------------

intersphinx_mapping = {
    "dask": ("https://docs.dask.org/en/stable/", None),
    "genno": ("https://genno.readthedocs.io/en/latest", None),
    "jpype": ("https://jpype.readthedocs.io/en/latest", None),
    "message_ix": ("https://docs.messageix.org/en/latest/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "pint": ("https://pint.readthedocs.io/en/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "sparse": ("https://sparse.pydata.org/en/stable/", None),
    "xarray": ("https://xarray.pydata.org/en/stable/", None),
}

# -- Options for sphinx.ext.todo -------------------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for sphinxcontrib.bibtext -------------------------------------------------

bibtex_bibfiles = ["references.bib"]



from pathlib import Path
from typing import Type

import click

import ixmp

ScenarioClass: Type[ixmp.Scenario] = ixmp.Scenario


class VersionType(click.ParamType):
    """A Click parameter type that accepts :class:`int` or 'all'."""

    name = "version"  # https://github.com/pallets/click/issues/411

    def convert(self, value, param, ctx):
        if value == "new":
            return value
        elif isinstance(value, int):
            return value
        else:
            try:
                return int(value)
            except ValueError:
                self.fail(f"{repr(value)} must be an integer or 'new'")


@click.group()
@click.option(
    "--url", metavar="ixmp://PLATFORM/MODEL/SCENARIO[#VERSION]", help="Scenario URL."
)
@click.option("--platform", help="Configured platform name.")
@click.option(
    "--dbprops",
    type=click.Path(exists=True, dir_okay=False),
    help="Database properties file.",
)
@click.option("--model", help="Model name.")
@click.option("--scenario", help="Scenario name.")
@click.option("--version", type=VersionType(), help="Scenario version.")
@click.pass_context
def main(ctx, url, platform, dbprops, model, scenario, version):
    # Load the indicated Platform
    mp = None
    if url:
        if dbprops or platform or model or scenario or version:
            raise click.UsageError(
                "--platform --model --scenario and/or " "--version redundant with --url"
            )

        scen, mp = ScenarioClass.from_url(url)
        ctx.obj = dict(scen=scen, mp=mp)
        return
    elif dbprops and platform:
        raise click.UsageError("give either --platform or --dbprops")
    elif platform:
        mp = ixmp.Platform(name=platform)
    elif dbprops:
        mp = ixmp.Platform(backend="jdbc", dbprops=dbprops)

    if not mp:
        return

    ctx.obj = dict(mp=mp)

    # Store the model and scenario name from arguments
    if model:
        ctx.obj["model name"] = model

    if scenario:
        ctx.obj["scenario name"] = scenario

    try:
        # Load the indicated Scenario
        if model and scenario:
            ctx.obj["scen"] = ScenarioClass(mp, model, scenario, version=version)
    except Exception as e:  # pragma: no cover
        raise click.ClickException(e.args[0])


@main.command()
@click.option("--config", help="Path to reporting configuration file.")
@click.argument("key")
@click.pass_obj
def report(context, config, key):
    """Run reporting for KEY."""
    # Import here to avoid importing reporting dependencies when running
    # other commands
    from ixmp.reporting import Reporter

    if not context:
        raise click.UsageError(
            "give either --url, --platform or --dbprops " "before command report"
        )

    # Instantiate the Reporter with the Scenario loaded by main()
    r = Reporter.from_scenario(context["scen"])

    # Read the configuration file, if any
    r.configure(config)

    # Print the target
    print(r.get(key).to_series().sort_index())


@main.command("show-versions")
def show_versions_cmd():
    """Print versions of ixmp and its dependencies."""
    ixmp.show_versions()


@main.command()
@click.option(
    "--remove-solution",
    is_flag=True,
    default=False,
    help="Forces removing solution if exists.",
)
@click.pass_obj
def solve(context, remove_solution):
    """Solve a Scenario and store results on the Platform.

    The scenario indicated by --url or --platform/--model/--scenario/--version is
    loaded, solved, and the solution results are saved on the Platform.

    If the scenario already has a solution, --remove-solution must be given.
    """
    if not context:
        raise click.UsageError("give --url before command solve")

    print("Run scenario solver")
    scen = context.get("scen")
    if not scen:
        print("Scenario not found")
        return
    if remove_solution and scen.has_solution():
        scen.remove_solution()
        print("Solution removed")
    scen.solve()
    print("Solver finished")


# "config" group


@main.group("config")
def config_group():
    """Get and set configuration keys."""


@config_group.command()
@click.argument("key", metavar="KEY")
def get(key):
    """Get configuration KEY."""
    print(ixmp.config.get(key))


@config_group.command()
def show():
    """Show all configuration."""
    print(
        f"Configuration path: {ixmp.config.path}",
        ixmp.config.path.read_text(),
        sep="\n\n",
    )


@config_group.command()
@click.argument("key", metavar="KEY")
@click.argument("value", nargs=-1)
def set(key, value):
    """Set configuration KEY to VALUE."""
    try:
        ixmp.config.set(key, value[0])
    except KeyError as e:
        raise click.ClickException(f"No registered configuration key {e}")
    else:
        # Save the configuration to file
        ixmp.config.save()
        print(f"Updated {ixmp.config.path}")


@main.command()
@click.option("--max-row", type=int, help="Max row numbers in each sheet.")
@click.argument("path", type=click.Path(writable=True))
@click.argument("filter_args", metavar="[--] FILTERS", nargs=-1)
@click.pass_obj
def export(context, path, filter_args, max_row):
    """Export scenario data to PATH.

    To export only part of the parameter data, e.g. for inspection, provide FILTERS in
    the format:

        … -- dim_1=val0,val1 dim_2=val2
    """
    # NB want to use type=click.Path(..., path_type=Path), but fails on bytes
    path = Path(path)

    if not context or "scen" not in context:
        raise click.UsageError(
            "give --url, or --platform, --model, and --scenario, before export"
        )

    # Convert additional arguments into a filters dict()
    filters = dict()
    for group in filter_args:
        dim, values = group.split("=", maxsplit=1)
        filters[dim] = list(values.split(","))

    context["scen"].to_excel(path, filters=filters, max_row=max_row)


@main.group("import")
@click.pass_obj
def import_group(context):
    """Import time series or scenario data.

    DATA is the path to a file containing input data in CSV (time series only) or Excel
    format.
    """
    if not context or "scen" not in context:
        raise click.UsageError(
            "give --url, or --platform, --model, and "
            "--scenario, before command import"
        )


@import_group.command("timeseries")
@click.option("--firstyear", type=int, help="First year of data to include.")
@click.option("--lastyear", type=int, help="Final year of data to include.")
@click.argument("file", type=click.Path(exists=True, dir_okay=False))
@click.pass_obj
def import_timeseries(context, file, firstyear, lastyear):
    """Import time series data."""
    context["scen"].read_file(Path(file), firstyear, lastyear)


@import_group.command("scenario")
@click.option(
    "--discard-solution", is_flag=True, help="Discard solution data if necessary."
)
@click.option("--add-units", is_flag=True, help="Add units to the Platform.")
@click.option("--init-items", is_flag=True, help="Initialize sets and parameters.")
@click.option("--commit-steps", is_flag=True, help="Commit after each step.")
@click.argument("file", type=click.Path(exists=True, dir_okay=False))
@click.pass_obj
def import_scenario(
    context, file, discard_solution, add_units, init_items, commit_steps
):
    """Import scenario data."""
    scenario = context["scen"]

    if scenario.has_solution() and discard_solution:
        scenario.remove_solution()

    try:
        scenario.check_out()
    except ValueError as e:
        raise click.ClickException(e.args[0])  # Show exception message to user
    except RuntimeError as e:
        if "not yet saved" in e.args[0]:
            pass  # --version=new; no need to check out
        else:  # pragma: no cover
            raise

    scenario.read_excel(
        Path(file),
        add_units=add_units,
        init_items=init_items,
        commit_steps=commit_steps,
    )


# "platform" group


@main.group("platform")
def platform_group():
    """Configure platforms and storage backends."""


@platform_group.command()
@click.argument("name", metavar="PLATFORM")
def remove(name):
    """Remove existing PLATFORM."""
    ixmp.config.remove_platform(name)
    print(f"Removed platform config for {repr(name)}")


@platform_group.command()
@click.argument("name", metavar="PLATFORM")
@click.argument("args", nargs=-1)
def add(name, args):
    """Add PLATFORM, configured with ARGS.

    If PLATFORM is 'default', ARGS must be the name of another platform.

    Otherwise, the first of ARGS is the backend. Currently only 'jdbc' is supported.
    For the 'jdbc' backend, the remaining ARGS must be one of:

    \b
    - oracle URL USERNAME PASSWORD: where the URL is something like
      example.com:PORT:SCHEMA. This configures a connection to an Oracle database.
    - hsqldb PATH: where PATH is the path to a file that will contain the database. To
      Use an existing database, give the file name without an extension.
    """
    ixmp.config.add_platform(name, *args)

    # Save the configuration to file
    ixmp.config.save()


@platform_group.command("list")
def list_platforms():
    """List configured platforms."""
    for key, info in ixmp.config.values["platform"].items():
        print(f"{key}: {info}")


@main.command("list")
@click.option(
    "--match",
    metavar="EXPR",
    default=None,
    help="Regular expression for model/scenario name.",
)
@click.option(
    "--default-only", is_flag=True, help="Only scenarios with a default version."
)
@click.option("--as-url", is_flag=True, help="Display outputs as ixmp URLs.")
@click.pass_obj
def list_scenarios(context, **kwargs):
    """List scenarios on the --platform."""
    from ixmp.utils import format_scenario_list

    if not context:
        raise click.UsageError(
            "give either --url, --platform or --dbprops " "before command list"
        )

    print(
        "\n".join(
            format_scenario_list(
                platform=context["mp"],
                model=context.get("model name", None),
                scenario=context.get("scenario name", None),
                **kwargs,
            )
        )
    )



from itertools import chain, repeat
from typing import List, Union

import dask
from genno.core.computer import Computer, Key

from ixmp.core import Scenario
from ixmp.reporting import computations
from ixmp.reporting.util import RENAME_DIMS, keys_for_quantity


class Reporter(Computer):
    """Class for describing and executing computations."""

    # Append ixmp.reporting.computations to the modules in which the Computer will look
    # up computations names.
    modules = list(Computer.modules) + [computations]

    @classmethod
    def from_scenario(cls, scenario: Scenario, **kwargs) -> "Reporter":
        """Create a Reporter by introspecting `scenario`.

        Parameters
        ----------
        scenario : .Scenario
            Scenario to introspect in creating the Reporter.
        kwargs : optional
            Passed to :meth:`Scenario.configure`.

        Returns
        -------
        .Reporter
            A Reporter instance containing:

            - A 'scenario' key referring to the *scenario* object.
            - Each parameter, equation, and variable in the *scenario*.
            - All possible aggregations across different sets of dimensions.
            - Each set in the *scenario*.
        """
        # New Reporter
        rep = cls(**kwargs)

        # Add the scenario itself
        rep.add("scenario", scenario)

        # List of top-level keys
        all_keys: List[Union[str, Key]] = []

        # List of parameters, equations, and variables
        quantities = chain(
            zip(repeat("par"), sorted(scenario.par_list())),
            zip(repeat("equ"), sorted(scenario.equ_list())),
            zip(repeat("var"), sorted(scenario.var_list())),
        )

        for ix_type, name in quantities:
            # List of computations for the quantity and maybe its marginals
            comps = keys_for_quantity(ix_type, name, scenario)

            # Add to the graph and index, including sums
            rep.add(*comps[0], strict=True, index=True, sums=True)

            try:
                # Add any marginals, but without sums
                rep.add(*comps[1], strict=True, index=True)
            except IndexError:
                pass  # Not an equ/var with marginals

            # Add keys to the list of all quantities
            all_keys.extend(c[0] for c in comps)

        # Add a key which simply collects all quantities
        rep.add("all", sorted(all_keys))

        # Add sets
        for name in scenario.set_list():
            elements = scenario.set(name)
            try:
                # Convert Series to list; protect list so that dask schedulers do not
                # try to interpret its contents as further tasks
                elements = dask.core.quote(elements.tolist())
            except AttributeError:  # pragma: no cover
                # pd.DataFrame for a multidimensional set; store as-is
                # TODO write tests for this
                pass

            rep.add(RENAME_DIMS.get(name, name), elements)

        return rep

    def finalize(self, scenario: Scenario) -> None:
        """Prepare the Reporter to act on `scenario`.

        The :class:`.TimeSeries` (i.e. including :class:`ixmp.Scenario` and
        :class:`message_ix.Scenario`) object `scenario` is stored with the key
        ``'scenario'``. All subsequent processing will act on data from this Scenario.
        """
        self.graph["scenario"] = scenario

    def set_filters(self, **filters) -> None:
        """Apply `filters` ex ante (before computations occur).

        See the description of :func:`.filters` under :ref:`reporting-config`.
        """
        self.configure(filters=filters)



from functools import lru_cache, partial
from typing import Dict

import pandas as pd
from genno import Key

#: Dimensions to rename when extracting raw data from Scenario objects.
#: Mapping from Scenario dimension name -> preferred dimension name.
RENAME_DIMS: Dict[str, str] = {}


def dims_for_qty(data):
    """Return the list of dimensions for *data*.

    If *data* is a :class:`pandas.DataFrame`, its columns are processed;
    otherwise it must be a list.

    genno.RENAME_DIMS is used to rename dimensions.
    """
    if isinstance(data, pd.DataFrame):
        # List of the dimensions
        dims = data.columns.tolist()
    else:
        dims = list(data)

    # Remove columns containing values or units; dimensions are the remainder
    for col in "value", "lvl", "mrg", "unit":
        try:
            dims.remove(col)
        except ValueError:
            continue

    # Rename dimensions
    return [RENAME_DIMS.get(d, d) for d in dims]


def keys_for_quantity(ix_type, name, scenario):
    """Return keys for *name* in *scenario*."""
    from .computations import data_for_quantity

    # Retrieve names of the indices of the ixmp item, without loading the data
    dims = dims_for_qty(scenario.idx_names(name))

    # Column for retrieving data
    column = "value" if ix_type == "par" else "lvl"

    # A computation to retrieve the data
    result = [
        (
            Key(name, dims),
            partial(data_for_quantity, ix_type, name, column),
            "scenario",
            "config",
        )
    ]

    # Add the marginal values at full resolution, but no aggregates
    if ix_type == "equ":
        result.append(
            (
                Key(f"{name}-margin", dims),
                partial(data_for_quantity, ix_type, name, "mrg"),
                "scenario",
                "config",
            )
        )

    return result


@lru_cache(1)
def get_reversed_rename_dims():
    return {v: k for k, v in RENAME_DIMS.items()}



from pathlib import Path

import pytest

from ixmp._config import KEYS, Config, _JSONEncoder, _locate


@pytest.fixture
def cfg():
    """Return a :class:`ixmp._config.Config` object without reading a file."""
    yield Config(read=False)


def test_encoder():
    # Custom encoder properly raises TypeError for unhandled value
    with pytest.raises(TypeError):
        _JSONEncoder().encode({"foo": range(10)})


def test_locate(cfg):
    try:
        # The result of this test depends on the user's environment. If
        # $HOME/.local/share/ixmp exists, the call will succeed; otherwise, it will
        # fail.
        _locate()
    except FileNotFoundError:  # pragma: no cover
        pass

    with pytest.raises(FileNotFoundError):
        _locate("nonexistent")


def test_set_get(cfg):
    # ixmp has no string keys by default, so we insert a fake one
    KEYS["test key"] = (str, None)
    cfg.values["test key"] = "foo"

    # get() works
    assert cfg.get("test key") == "foo"

    # set() changes the value
    cfg.set("test key", "bar")
    assert cfg.get("test key") == "bar"

    # set() with None makes no change
    cfg.set("test key", None)
    assert cfg.get("test key") == "bar"

    # set() with invalid type raises exception
    KEYS["test key"] = (int, None)
    with pytest.raises(TypeError):
        cfg.set("test key", "foo")


def test_register(cfg):
    # New key can be registered
    cfg.register("new key", int, 42)

    # Default value is set on the instance
    assert cfg.get("new key") == 42

    # Can't re-register an existing key
    with pytest.raises(KeyError):
        cfg.register("new key", int, 43)

    # Register a key with type Path
    cfg.register("new key 2", Path)

    # The key can be with a string value, automatically converted to Path
    cfg.set("new key 2", "/foo/bar/baz")
    assert isinstance(cfg.get("new key 2"), Path)


def test_config_platform(cfg):
    # Default platform is 'local'
    assert cfg.values["platform"]["default"] == "local"

    # Set another platform as the default
    cfg.add_platform("dummy", "jdbc", "oracle", "url", "user", "password")
    cfg.add_platform("default", "dummy")

    # Now the default platform is 'dummy'
    assert cfg.values["platform"]["default"] == "dummy"

    # Same info is retrieved for 'default' and the actual name
    assert cfg.get_platform_info("default") == cfg.get_platform_info("dummy")

    # Invalid calls
    with pytest.raises(ValueError):
        cfg.add_platform("invalid", "notabackend", "other", "args")

    with pytest.raises(ValueError):
        cfg.get_platform_info("nonexistent")



import logging
import sys
from subprocess import Popen
from time import sleep

import pytest
from pretenders.client.http import HTTPMock
from pretenders.common.constants import FOREVER

import ixmp
from ixmp.testing import create_test_platform

log = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def server():
    proc = Popen(
        [
            sys.executable,
            "-m",
            "pretenders.server.server",
            "--host",
            "localhost",
            "--port",
            "8000",
        ]
    )
    log.info(f"Mock server started with pid {proc.pid}")

    # Wait for server to start up
    sleep(5)

    yield

    proc.terminate()
    log.info("Mock server terminated")


@pytest.fixture(scope="function")
def mock(server):
    # Create the mock server
    httpmock = HTTPMock("localhost", 8000)

    # Common responses for both tests
    httpmock.when("POST /login").reply(
        '"security-token"', headers={"Content-Type": "application/json"}, times=FOREVER
    )

    yield httpmock


def test_check_single_model_access(mock, tmp_path, test_data_path):
    mock.when(
        "POST /access/list",
        body='.+"test_user".+',
        headers={"Authorization": "Bearer security-token"},
    ).reply("[true]", headers={"Content-Type": "application/json"}, times=FOREVER)
    mock.when(
        "POST /access/list",
        body='.+"non_granted_user".+',
        headers={"Authorization": "Bearer security-token"},
    ).reply("[false]", headers={"Content-Type": "application/json"}, times=FOREVER)

    test_props = create_test_platform(
        tmp_path, test_data_path, "access", auth_url=mock.pretend_url
    )

    mp = ixmp.Platform(backend="jdbc", dbprops=test_props)

    granted = mp.check_access("test_user", "test_model")
    assert granted

    granted = mp.check_access("non_granted_user", "test_model")
    assert not granted

    granted = mp.check_access("non_existing_user", "test_model")
    assert not granted


def test_check_multi_model_access(mock, tmp_path, test_data_path):
    mock.when(
        "POST /access/list",
        body='.+"test_user".+',
        headers={"Authorization": "Bearer security-token"},
    ).reply(
        "[true, false]", headers={"Content-Type": "application/json"}, times=FOREVER
    )
    mock.when(
        "POST /access/list",
        body='.+"non_granted_user".+',
        headers={"Authorization": "Bearer security-token"},
    ).reply(
        "[false, false]", headers={"Content-Type": "application/json"}, times=FOREVER
    )

    test_props = create_test_platform(
        tmp_path, test_data_path, "access", auth_url=mock.pretend_url
    )

    mp = ixmp.Platform(backend="jdbc", dbprops=test_props)

    access = mp.check_access("test_user", ["test_model", "non_existing_model"])
    assert access["test_model"]
    assert not access["non_existing_model"]

    access = mp.check_access("non_granted_user", ["test_model", "non_existing_model"])
    assert not access["test_model"]
    assert not access["non_existing_model"]

    access = mp.check_access("non_existing_user", ["test_model", "non_existing_model"])
    assert not access["test_model"]
    assert not access["non_existing_model"]



import logging

import numpy as np
import pytest
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

import ixmp
from ixmp.testing import HIST_DF, TS_DF, make_dantzig, models

TS_DF_CLEARED = TS_DF.copy()
TS_DF_CLEARED.loc[0, 2005] = np.nan


def test_run_clone(caplog, test_mp):
    caplog.set_level(logging.WARNING)

    # this test is designed to cover the full functionality of the GAMS API
    # - initialize a new platform instance
    # - creates a new scenario and exports a gdx file
    # - runs the tutorial transport model
    # - reads back the solution from the output
    # - performs the test on the objective value and the timeseries data
    mp = test_mp
    scen = make_dantzig(mp, solve=True, quiet=True)
    assert np.isclose(scen.var("z")["lvl"], 153.675)
    assert_frame_equal(scen.timeseries(iamc=True), TS_DF)

    # cloning with `keep_solution=True` keeps all timeseries and the solution
    scen2 = scen.clone(keep_solution=True)
    assert np.isclose(scen2.var("z")["lvl"], 153.675)
    assert_frame_equal(scen2.timeseries(iamc=True), TS_DF)

    # version attribute of the clone increments the original (GitHub #211)
    assert scen2.version == scen.version + 1

    # cloning with `keep_solution=True` and `first_model_year` raises a warning
    scen.clone(keep_solution=True, shift_first_model_year=2005)
    assert (
        "Override keep_solution=True for shift_first_model_year"
        == caplog.records[-1].message
    )

    # cloning with `keep_solution=False` drops the solution and only keeps
    # timeseries set as `meta=True`
    scen3 = scen.clone(keep_solution=False)
    assert np.isnan(scen3.var("z")["lvl"])
    assert_frame_equal(scen3.timeseries(iamc=True), HIST_DF)

    # cloning with `keep_solution=False` and `first_model_year`
    # drops the solution and removes all timeseries not marked `meta=True`
    # in the model horizon (i.e, `year >= first_model_year`)
    scen4 = scen.clone(keep_solution=False, shift_first_model_year=2005)
    assert np.isnan(scen4.var("z")["lvl"])
    assert_frame_equal(scen4.timeseries(iamc=True), TS_DF_CLEARED)


def test_run_remove_solution(test_mp):
    # create a new instance of the transport problem and solve it
    mp = test_mp
    scen = make_dantzig(mp, solve=True, quiet=True)
    assert np.isclose(scen.var("z")["lvl"], 153.675)

    # check that re-solving the model will raise an error if a solution exists
    pytest.raises(ValueError, scen.solve, model="dantzig", case="fail")

    # remove the solution, check that variables are empty
    # and timeseries not marked `meta=True` are removed
    scen2 = scen.clone()
    scen2.remove_solution()
    assert not scen2.has_solution()
    assert np.isnan(scen2.var("z")["lvl"])
    assert_frame_equal(scen2.timeseries(iamc=True), HIST_DF)

    # remove the solution with a specific year as first model year, check that
    # variables are empty and timeseries not marked `meta=True` are removed
    scen3 = scen.clone()
    scen3.remove_solution(first_model_year=2005)
    assert not scen3.has_solution()
    assert np.isnan(scen3.var("z")["lvl"])
    assert_frame_equal(scen3.timeseries(iamc=True), TS_DF_CLEARED)


def scenario_list(mp):
    return mp.scenario_list(default=False)[["model", "scenario"]]


def assert_multi_db(mp1, mp2):
    assert_frame_equal(scenario_list(mp1), scenario_list(mp2))


def get_distance(scen):
    return scen.par("d").set_index(["i", "j"]).loc["san-diego", "topeka"]["value"]


def test_multi_db_run(tmpdir):
    # create a new instance of the transport problem and solve it
    mp1 = ixmp.Platform(backend="jdbc", driver="hsqldb", path=tmpdir / "mp1")
    scen1 = make_dantzig(mp1, solve=True, quiet=True)

    mp2 = ixmp.Platform(backend="jdbc", driver="hsqldb", path=tmpdir / "mp2")
    # add other unit to make sure that the mapping is correct during clone
    mp2.add_unit("wrong_unit")
    mp2.add_region("wrong_region", "country")

    # check that cloning across platforms must copy the full solution
    pytest.raises(NotImplementedError, scen1.clone, platform=mp2, keep_solution=False)

    # clone solved model across platforms (with default settings)
    scen1.clone(platform=mp2, keep_solution=True)

    # close the db to ensure that data and solution of the clone are saved
    mp2.close_db()
    del mp2

    # reopen the connection to the second platform and reload scenario
    _mp2 = ixmp.Platform(backend="jdbc", driver="hsqldb", path=tmpdir / "mp2")
    assert_multi_db(mp1, _mp2)
    scen2 = ixmp.Scenario(_mp2, **models["dantzig"])

    # check that sets, variables and parameter were copied correctly
    assert_array_equal(scen1.set("i"), scen2.set("i"))
    assert_frame_equal(scen1.par("d"), scen2.par("d"))
    assert np.isclose(scen2.var("z")["lvl"], 153.675)
    assert_frame_equal(scen1.var("x"), scen2.var("x"))

    # check that custom unit, region and timeseries are migrated correctly
    assert scen2.par("f")["value"] == 90.0
    assert scen2.par("f")["unit"] == "USD/km"
    assert_frame_equal(scen2.timeseries(iamc=True), TS_DF)


def test_multi_db_edit_source(tmpdir):
    # create a new instance of the transport problem
    mp1 = ixmp.Platform(backend="jdbc", driver="hsqldb", path=tmpdir / "mp1")
    scen1 = make_dantzig(mp1)

    mp2 = ixmp.Platform(backend="jdbc", driver="hsqldb", path=tmpdir / "mp2")
    scen2 = scen1.clone(platform=mp2)

    assert_frame_equal(scen1.par("d"), scen2.par("d"))

    scen1.check_out()
    scen1.add_par("d", ["san-diego", "topeka"], 1.5, "km")
    scen1.commit("foo")

    obs = get_distance(scen1)
    exp = 1.5
    assert np.isclose(obs, exp)

    obs = get_distance(scen2)
    exp = 1.4
    assert np.isclose(obs, exp)

    assert_multi_db(mp1, mp2)


def test_multi_db_edit_target(tmpdir):
    # create a new instance of the transport problem
    mp1 = ixmp.Platform(backend="jdbc", driver="hsqldb", path=tmpdir / "mp1")
    scen1 = make_dantzig(mp1)

    mp2 = ixmp.Platform(backend="jdbc", driver="hsqldb", path=tmpdir / "mp2")
    scen2 = scen1.clone(platform=mp2)

    assert_frame_equal(scen1.par("d"), scen2.par("d"))

    scen2.check_out()
    scen2.add_par("d", ["san-diego", "topeka"], 1.5, "km")
    scen2.commit("foo")

    obs = get_distance(scen2)
    exp = 1.5
    assert np.isclose(obs, exp)

    obs = get_distance(scen1)
    exp = 1.4
    assert np.isclose(obs, exp)

    assert_multi_db(mp1, mp2)



"""Tests for compatibility with other packages, especially :mod:`message_ix`."""
from pathlib import Path

import pytest

from ixmp import _config, config


@pytest.fixture()
def tmp_path(tmp_path_factory):
    p0 = tmp_path_factory.mktemp("foo")
    p1 = p0.joinpath("bar", "baz")
    p1.mkdir(exist_ok=True, parents=True)
    yield p1


def test_message_model_dir(ixmp_cli, tmp_env, tmp_path):
    """Configuration keys like 'message model dir' can be registered, set, and used."""
    value = tmp_path

    key = "TEST message model dir"

    # Setting the key without first registering it results in an error
    result = ixmp_cli.invoke(["config", "set", key, str(value)])
    assert 1 == result.exit_code
    assert (
        f"Error: No registered configuration key {repr(key)}" == result.output.strip()
    )

    # Register the key
    config.register(key, Path, Path(__file__).parent / "model")

    # The key can be set
    result = ixmp_cli.invoke(["config", "set", key, str(value)])
    assert 0 == result.exit_code

    # Force re-read of the key from file
    config.read()

    # Configuration value has the expected type
    assert isinstance(config.get(key), Path)
    # Expected value
    assert value == config.get(key)

    # Configuration
    result = ixmp_cli.invoke(["config", "get", key])
    assert str(value) == result.output.strip()

    # Clean for remainder of tests
    config.unregister(key)
    assert key not in _config.KEYS



import logging
import re
import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Dict, Iterator, List, Tuple
from urllib.parse import urlparse
from warnings import warn

import pandas as pd

from ixmp.backend import ItemType

log = logging.getLogger(__name__)

# Globally accessible logger.
# TODO remove when :func:`logger` is removed.
_LOGGER = None


def logger():
    """Access global logger.

    .. deprecated:: 3.3
       To control logging from ixmp, instead use :mod:`logging` to retrieve it:

       .. code-block:: python

          import logging
          ixmp_logger = logging.getLogger("ixmp")

          # Example: set the level to INFO
          ixmp_logger.setLevel(logging.INFO)
    """
    warn(
        "ixmp.utils.logger() is deprecated as of 3.3.0, and will be removed in ixmp "
        '5.0. Use logging.getLogger("ixmp").',
        DeprecationWarning,
    )
    return logging.getLogger("ixmp")


def as_str_list(arg, idx_names=None):
    """Convert various *arg* to list of str.

    Several types of arguments are handled:
    - None: returned as None.
    - str: returned as a length-1 list of str.
    - iterable of values: returned as a list with each value converted to str
    - dict, with list of idx_names: the idx_names are used to look up values
      in the dict, the resulting list has the corresponding values in the same
      order.

    """
    if arg is None:
        return None
    elif idx_names is None:
        # arg must be iterable
        # NB narrower ABC Sequence does not work here; e.g. test_excel_io()
        #    fails via Scenario.add_set().
        if isinstance(arg, Iterable) and not isinstance(arg, str):
            return list(map(str, arg))
        else:
            return [str(arg)]
    else:
        return [str(arg[idx]) for idx in idx_names]


def isscalar(x):
    """Returns True if x is a scalar"""
    return not isinstance(x, Iterable) or isinstance(x, str)


def check_year(y, s):
    """Returns True if y is an int, raises an error if y is not None"""
    if y is not None:
        if not isinstance(y, int):
            raise ValueError("arg `{}` must be an integer!".format(s))
        return True


def diff(a, b, filters=None) -> Iterator[Tuple[str, pd.DataFrame]]:
    """Compute the difference between Scenarios `a` and `b`.

    :func:`diff` combines :func:`pandas.merge` and :meth:`Scenario.items`.
    Only parameters are compared. :func:`~pandas.merge` is called with the
    arguments ``how="outer", sort=True, suffixes=("_a", "_b"), indicator=True";
    the merge is performed on all columns except 'value' or 'unit'.

    Yields
    ------
    tuple of str, pandas.DataFrame
        Tuples of item name and data.
    """
    # Iterators; index 0 corresponds to `a`, 1 to `b`
    items = [
        a.items(filters=filters, type=ItemType.PAR),
        b.items(filters=filters, type=ItemType.PAR),
    ]
    # State variables for loop
    name = ["", ""]
    data: List[pd.DataFrame] = [None, None]

    # Elements for first iteration
    name[0], data[0] = next(items[0])
    name[1], data[1] = next(items[1])

    while True:
        # Convert scalars to pd.DataFrame
        data = list(map(maybe_convert_scalar, data))

        # Compare the names from `a` and `b` to ensure matching items
        if name[0] == name[1]:
            # Matching items in `a` and `b`
            current_name = name[0]
            left, right = data
        else:
            # Mismatched; either `a` or `b` has no data for these filters
            current_name = min(*name)
            if name[0] > current_name:
                # No data in `a` for `current_name`; create an empty DataFrame
                left, right = pd.DataFrame(columns=data[1].columns), data[1]
            else:
                left, right = data[0], pd.DataFrame(columns=data[0].columns)

        # Either merge on remaining columns; or, for scalars, on the indices
        on = sorted(set(left.columns) - {"value", "unit"})
        on_arg: Dict[str, object] = (
            dict(on=on) if on else dict(left_index=True, right_index=True)
        )

        # Merge the data from each side
        yield current_name, pd.merge(
            left,
            right,
            how="outer",
            **on_arg,
            sort=True,
            suffixes=("_a", "_b"),
            indicator=True,
        )

        # Maybe advance each iterators
        for i in (0, 1):
            try:
                if name[i] == current_name:
                    # data was compared in this iteration; advance
                    name[i], data[i] = next(items[i])
            except StopIteration:
                # No more data for this iterator.
                # Use "~" because it sorts after all ASCII characters
                name[i], data[i] = "~ end", None

        if name[0] == name[1] == "~ end":
            break


def maybe_check_out(timeseries, state=None):
    """Check out `timeseries` depending on `state`.

    If `state` is :obj:`None`, then :meth:`check_out` is called.

    Returns
    -------
    :obj:`True`
        if `state` was :obj:`None` and a check out was performed, i.e.
        `timeseries` was previously in a checked-in state.
    :obj:`False`
        if `state` was :obj:`None` and no check out was performed, i.e.
        `timeseries` was already in a checked-out state.
    `state`
        if `state` was not :obj:`None` and no check out was attempted.

    Raises
    ------
    ValueError
        If `timeseries` is a :class:`.Scenario` object and
        :meth:`~.Scenario.has_solution` is :obj:`True`.

    See Also
    --------
    :meth:`.TimeSeries.check_out`
    :meth:`.Scenario.check_out`
    """
    if state is not None:
        return state

    try:
        timeseries.check_out()
    except RuntimeError:
        # If `timeseries` is new (has not been committed), the checkout
        # attempt raises an exception
        return False
    else:
        return True


def maybe_commit(timeseries, condition, message):
    """Commit `timeseries` with `message` if `condition` is :obj:`True`.

    Returns
    -------
    :obj:`True`
        if a commit is performed.
    :obj:`False`
        if any exception is raised during the attempted commit. The exception
        is logged with level ``INFO``.

    See Also
    --------
    :meth:`.TimeSeries.commit`
    """
    if not condition:
        return False

    try:
        timeseries.commit(message)
    except RuntimeError as exc:
        log.info(f"maybe_commit() didn't commit: {exc}")
        return False
    else:
        return True


def maybe_convert_scalar(obj) -> pd.DataFrame:
    """Convert `obj` to :class:`pandas.DataFrame`.

    Parameters
    ----------
    obj
        Any value returned by :meth:`Scenario.par`. For a scalar
        (0-dimensional) parameter, this will be :class:`dict`.

    Returns
    -------
    pandas.DataFrame
        :meth:`maybe_convert_scalar` always returns a data frame.
    """
    if isinstance(obj, dict):
        return pd.DataFrame.from_dict({0: obj}, orient="index")
    else:
        return obj


def parse_url(url):
    """Parse *url* and return Platform and Scenario information.

    A URL (Uniform Resource Locator), as the name implies, uniquely identifies
    a specific scenario and (optionally) version of a model, as well as
    (optionally) the database in which it is stored. ixmp URLs take forms
    like::

        ixmp://PLATFORM/MODEL/SCENARIO[#VERSION]
        MODEL/SCENARIO[#VERSION]

    where:

    - The ``PLATFORM`` is a configured platform name; see :obj:`ixmp.config`.
    - ``MODEL`` may not contain the forward slash character ('/'); ``SCENARIO``
      may contain any number of forward slashes. Both must be supplied.
    - ``VERSION`` is optional but, if supplied, must be an integer.

    Returns
    -------
    platform_info : dict
        Keyword argument 'name' for the :class:`Platform` constructor.
    scenario_info : dict
        Keyword arguments for a :class:`Scenario` on the above platform:
        'model', 'scenario' and, optionally, 'version'.

    Raises
    ------
    ValueError
        For malformed URLs.
    """
    components = urlparse(url)
    if components.scheme not in ("ixmp", ""):
        raise ValueError("URL must begin with ixmp:// or //")

    platform_info = dict()
    if components.netloc:
        platform_info["name"] = components.netloc

    scenario_info = dict()

    path = components.path.split("/")
    if len(path):
        # If netloc was given, path begins with '/'; discard
        path = path if len(path[0]) else path[1:]

        if len(path) < 2:
            raise ValueError("URL path must be 'MODEL/SCENARIO'")

        scenario_info["model"] = path[0]
        scenario_info["scenario"] = "/".join(path[1:])

    if len(components.query):
        raise ValueError(f"queries ({components.query}) not supported in URLs")

    if len(components.fragment):
        try:
            version = int(components.fragment)
        except ValueError:
            if components.fragment != "new":
                raise
            else:
                version = "new"
        finally:
            scenario_info["version"] = version

    return platform_info, scenario_info


def year_list(x):
    """Return the elements of x that can be cast to year (int)."""
    lst = []
    for i in x:
        try:
            int(i)  # this is a year
            lst.append(i)
        except ValueError:
            pass
    return lst


def filtered(df, filters):
    """Returns a filtered dataframe based on a filters dictionary"""
    if filters is None:
        return df

    mask = pd.Series(True, index=df.index)
    for k, v in filters.items():
        isin = df[k].isin(as_str_list(v))
        mask = mask & isin
    return df[mask]


def format_scenario_list(
    platform, model=None, scenario=None, match=None, default_only=False, as_url=False
):
    """Return a formatted list of TimeSeries on *platform*.

    Parameters
    ----------
    platform : :class:`.Platform`
    model : str, optional
        Model name to restrict results. Passed to :meth:`.scenario_list`.
    scenario : str, optional
        Scenario name to restrict results. Passed to :meth:`.scenario_list`.
    match : str, optional
        Regular expression to restrict results. Only results where the model or
        scenario name matches are returned.
    default_only : bool, optional
        Only return TimeSeries where a default version has been set with
        :meth:`TimeSeries.set_as_default`.
    as_url : bool, optional
        Format results as ixmp URLs.

    Returns
    -------
    list of str
        If *as_url* is :obj:`False`, also include summary information.
    """

    try:
        match = re.compile(".*" + match + ".*")
    except TypeError:
        pass

    def describe(df):
        N = len(df)
        min = df.version.min()
        max = df.version.max()

        result = dict(N=N, range="")
        if N > 1:
            result["range"] = "{}–{}".format(min, max)
            if N != max:
                result["range"] += " ({} versions)".format(N)

        try:
            mask = df.is_default.astype(bool)
            result["default"] = df.loc[mask, "version"].iat[0]
        except IndexError:
            result["default"] = max

        return pd.Series(result)

    info = (
        platform.scenario_list(model=model, scen=scenario, default=default_only)
        .groupby(["model", "scenario"])
        .apply(describe)
        .reset_index()
    )

    if not len(info):
        # No results; re-create a minimal empty data frame
        info = pd.DataFrame([], columns=["model", "scenario", "default", "N"])

    info["scenario"] = info["scenario"].str.cat(info["default"].astype(str), sep="#")

    if match:
        info = info[info["model"].str.match(match) | info["scenario"].str.match(match)]

    lines = []

    if as_url:
        info["url"] = "ixmp://{}".format(platform.name)
        urls = info["url"].str.cat([info["model"], info["scenario"]], sep="/")
        lines = urls.tolist()
    else:
        width = 0 if not len(info) else info["scenario"].str.len().max()
        info["scenario"] = info["scenario"].str.ljust(width)

        for model, m_info in info.groupby(["model"]):
            lines.extend(
                [
                    "",
                    model + "/",
                    "  " + "\n  ".join(m_info["scenario"].str.cat(m_info["range"])),
                ]
            )

        lines.append("")

    # Summary information
    if not as_url:
        lines.extend(
            [
                str(len(info["model"].unique())) + " model name(s)",
                str(len(info["scenario"].unique())) + " scenario name(s)",
                str(len(info)) + " (model, scenario) combination(s)",
                str(info["N"].sum()) + " total scenarios",
            ]
        )

    return lines


def show_versions(file=sys.stdout):
    """Print information about ixmp and its dependencies to *file*."""
    import importlib
    from subprocess import DEVNULL, CalledProcessError, check_output

    from xarray.util.print_versions import get_sys_info

    from ixmp.model.gams import gams_version

    def _git_log(mod):
        cmd = ["git", "log", "--oneline", "--no-color", "--decorate", "-n 1"]
        try:
            cwd = Path(mod.__file__).parent
            info = check_output(cmd, cwd=cwd, encoding="utf-8", stderr=DEVNULL)
        except Exception:
            # Occurs if "git log" fails; or if mod.__file__ is None (#338)
            return ""
        else:
            return f"\n     {info.rstrip()}"

    deps = [
        None,  # Prints a separator
        # ixmp stack
        "ixmp",
        "message_ix",
        "message_ix_models",
        "message_data",
        None,
        # ixmp dependencies
        "click",
        "dask",
        "genno",
        "graphviz",
        "jpype",
        "openpyxl",
        "pandas",
        "pint",
        "xarray",
        "yaml",
        None,
        # Optional dependencies, dependencies of message_ix and message_data
        "iam_units",
        "jupyter",
        "matplotlib",
        "plotnine",
        "pyam",
        None,
    ]

    info = []
    for module_name in deps:
        try:
            # Import the module
            mod = sys.modules.get(module_name, None) or importlib.import_module(
                module_name
            )
        except Exception:
            # Couldn't import
            info.append((module_name, None))
            continue

        # Retrieve git log information, if any
        gl = _git_log(mod)
        try:
            version = mod.__version__
        except Exception:
            # __version__ not available
            version = "installed"
        finally:
            info.append((module_name, version + gl))

        if module_name == "jpype":
            info.append(("… JVM path", mod.getDefaultJVMPath()))

    # Also display GAMS version, if any
    try:
        version = gams_version()
    except (CalledProcessError, FileNotFoundError):
        version = "'gams' executable not in PATH"
    finally:
        info.extend([("GAMS", version), (None, None)])

    # Use xarray to get system & Python information
    info.extend(get_sys_info()[1:])  # Exclude the commit number

    for k, stat in info:
        if (k, stat) == (None, None):
            # Separator line
            print("", file=file)
        else:
            print(f"{k + ':':12} {stat}", file=file)


def update_par(scenario, name, data):
    """Update parameter *name* in *scenario* using *data*, without overwriting.

    Only values which do not already appear in the parameter data are added.
    """
    tmp = pd.concat([scenario.par(name), data])
    columns = list(filter(lambda c: c != "value", tmp.columns))
    tmp = tmp.drop_duplicates(subset=columns, keep=False)

    if len(tmp):
        scenario.add_par(name, tmp)
